# -*- coding: utf-8 -*-
#
# The Selector is a Python library of algorithms for selecting diverse
# subsets of data for machine-learning.
#
# Copyright (C) 2022-2024 The QC-Devs Community
#
# This file is part of Selector.
#
# Selector is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Selector is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Module for Distance-Based Selection Methods."""

import bitarray
import numpy as np
from scipy import spatial
from typing import List, Iterable, Union

from selector.methods.base import SelectionBase
from selector.methods.utils import optimize_radius

__all__ = [
    "MaxMin",
    "MaxSum",
    "OptiSim",
    "DISE",
]


class MaxMin(SelectionBase):
    """Select samples using MaxMin algorithm.

    MaxMin is possibly the most widely used method for dissimilarity-based
    compound selection. When presented with a dataset of samples, the
    initial point is chosen as the dataset's medoid center. Next, the second
    point is chosen to be that which is furthest from this initial point.
    Subsequently, all following points are selected via the following
    logic:

    1. Find the minimum distance from every point to the already-selected ones.
    2. Select the point which has the maximum distance among those calculated
       in the previous step.

    In the current implementation, this method requires or computes the full pairwise-distance
    matrix, so it is not recommended for large datasets.

    References
    ----------
    [1] Ashton, Mark, et al., Identification of diverse database subsets using
    property‐based and fragment‐based molecular descriptions, Quantitative
    Structure‐Activity Relationships 21.6 (2002): 598-604.
    """

    def __init__(self, fun_dist=None, ref_index=None):
        """
        Initializing class.

        Parameters
        ----------
        fun_distance : callable
            Function for calculating the pairwise distance between sample points.
            `fun_dist(x) -> x_dist` takes a 2D feature array of shape (n_samples, n_features)
            and returns a 2D distance array of shape (n_samples, n_samples).
        ref_index: int, list, optional
            Index of the reference sample to start the selection algorithm from.
            It can be an integer, or a list of integers or None. When None, the medoid center is chosen as the reference
            sample.
            When the `ref_index` is a list for multiple classes, it will be shared among all clusters.
            If we want to use different reference indices for each class, we can perform the subset
            selection for each class separately where different `ref_index` parameters can be used.
            For example, if we have two classes, we can pass `ref_index=[0, 1]` to select samples
            from class 0 and `ref_index=[3, 6]` class 1 respectively.

        """
        self.fun_dist = fun_dist
        self.ref_index = ref_index

    def select_from_cluster(self, x, size, labels=None) -> Union[List, Iterable]:
        """Return selected samples from a cluster based on MaxMin algorithm.

        Parameters
        ----------
        x: ndarray of shape (n_samples, n_features) or (n_samples, n_samples)
            Feature matrix of `n_samples` samples in `n_features` dimensional feature space,
            or the pairwise distance matrix between `n_samples` samples.
            If `fun_dist` is `None`, the `x` is assumed to be a square pairwise distance matrix.
        size: int
            Number of sample points to select (i.e. size of the subset).
        labels: np.ndarray
            Indices of samples that form a cluster.

        Returns
        -------
        selected : Union[List, Iterable]
            List of indices of selected samples.
        """
        # calculate pairwise distance between points
        x_dist = x
        if self.fun_dist is not None:
            x_dist = self.fun_dist(x)
        # check x_dist is a square symmetric matrix
        if x_dist.shape[0] != x_dist.shape[1]:
            raise ValueError(f"The pairwise distance matrix must be square, got {x_dist.shape}.")
        if np.max(abs(x_dist - x_dist.T)) > 1e-8:
            raise ValueError("The pairwise distance matrix must be symmetric.")

        if labels is not None:
            # extract pairwise distances from full pairwise distance matrix to obtain a new matrix
            # that only contains pairwise distances between samples within a given cluster
            x_dist = x_dist[labels][:, labels]

        # choosing initial point
        selected = get_initial_selection(
            x=None, x_dist=x_dist, ref_index=self.ref_index, fun_dist=None
        )

        # select following points until desired number of points have been obtained
        while len(selected) < size:
            # determine the min pairwise distances between the selected points and all other points
            min_distances = np.min(x_dist[selected], axis=0)
            # determine which point affords the maximum distance among the minimum distances
            # captured in min_distances
            new_id = np.argmax(min_distances)
            selected.append(new_id)

        selected = [int(i) for i in selected]

        return selected


class MaxSum(SelectionBase):
    """Select samples using MaxSum algorithm.

    Whereas the goal of the MaxMin algorithm is to maximize the minimum distance
    between any pair of distinct elements in the selected subset of a dataset,
    the MaxSum algorithm aims to maximize the sum of distances between all
    pairs of elements in the selected subset. When presented with a dataset of
    samples, the initial point is chosen as the dataset's medoid center. Next,
    the second point is chosen to be that which is furthest from this initial
    point. Subsequently, all following points are selected via the following
    logic:

    1. Determine the sum of distances from every point to the already-selected ones.
    2. Select the point which has the maximum sum of distances among those calculated
       in the previous step.

    References
    ----------
    [1] Borodin, Allan, Hyun Chul Lee, and Yuli Ye, Max-sum diversification, monotone
    submodular functions and dynamic updates, Proceedings of the 31st ACM SIGMOD-SIGACT-SIGAI
    symposium on Principles of Database Systems. 2012.
    """

    def __init__(self, fun_dist=None, ref_index=None):
        """
        Initializing class.

        Parameters
        ----------
        fun_dist : callable
            Function for calculating the pairwise distance between sample points.
            `fun_dist(x) -> x_dist` takes a 2D feature array of shape (n_samples, n_features)
            and returns a 2D distance array of shape (n_samples, n_samples).
        ref_index: int, list, optional
            Index of the reference sample to start the selection algorithm from.
            It can be an integer, or a list of integers or None. When None, the medoid center is chosen as the reference
            sample.
            When the `ref_index` is a list for multiple classes, it will be shared among all clusters.
            If we want to use different reference indices for each class, we can perform the subset
            selection for each class separately where different `ref_index` parameters can be used.
            For example, if we have two classes, we can pass `ref_index=[0, 1]` to select samples
            from class 0 and `ref_index=[3, 6]` class 1 respectively.

        """
        self.fun_dist = fun_dist
        self.ref_index = ref_index

    def select_from_cluster(self, x, size, labels=None) -> Union[List, Iterable]:
        """Return selected samples from a cluster based on MaxSum algorithm.

        Parameters
        ----------
        x: ndarray of shape (n_samples, n_features) or (n_samples, n_samples)
            Feature matrix of `n_samples` samples in `n_features` dimensional feature space,
            or the pairwise distance matrix between `n_samples` samples.
            If `fun_dist` is `None`, the `x` is assumed to be a square pairwise distance matrix.
        size: int
            Number of sample points to select (i.e. size of the subset).
        labels: np.ndarray
            Indices of samples that form a cluster.

        Returns
        -------
        selected : Union[List, Iterable]
            List of indices of selected samples.

        """
        # calculate pairwise distance between points
        x_dist = x
        if self.fun_dist is not None:
            x_dist = self.fun_dist(x)
        # check x_dist is a square symmetric matrix
        if x_dist.shape[0] != x_dist.shape[1]:
            raise ValueError(f"The pairwise distance matrix must be square, got {x_dist.shape}.")
        if np.max(abs(x_dist - x_dist.T)) > 1e-8:
            raise ValueError("The pairwise distance matrix must be symmetric.")

        if labels is not None:
            # extract pairwise distances from full pairwise distance matrix to obtain a new matrix
            # that only contains pairwise distances between samples within a given cluster.
            x_dist = x_dist[labels][:, labels]

        # setting up initial point
        selected = get_initial_selection(
            x=None, x_dist=x_dist, ref_index=self.ref_index, fun_dist=None
        )
        # select following points until desired number of points have been obtained
        while len(selected) < size:
            # determine sum of pairwise distances between selected points and all other points
            sum_distances = np.sum(x_dist[selected], axis=0)
            # determine which point has the max sum of pairwise distances to already-selected points
            new_id = np.argmax(sum_distances)
            # make sure that new_id corresponds to a new point
            while new_id in selected:
                # set the sum of distances for the current point corresponding to new_id to 0
                sum_distances[new_id] = 0
                # find a different point with the maximum sum of pairwise distances to
                # already-selected points
                new_id = np.argmax(sum_distances)
            selected.append(new_id)

        selected = [int(i) for i in selected]
        return selected


class OptiSim(SelectionBase):
    """Selecting samples using OptiSim algorithm.

    The OptiSim algorithm selects samples from a dataset by first choosing the medoid center as the
    initial point. Next, points are randomly chosen and added to a subsample if they exist
    outside of radius r from all previously selected points (otherwise, they are discarded). Once k
    number of points have been added to the subsample, the point with the greatest minimum distance
    to the previously selected points is chosen. Then, the subsample is cleared and the process is
    repeated.

    Notes
    -----
    When the `ref_index` is a list for multiple classes, it will be shared among all clusters.
    If we want to use different reference indices for each class, we can perform the subset
    selection for each class separately where different `ref_index` parameters can be used.
    For example, if we have two classes, we can pass `ref_index=[0, 1]` to select samples from
    class 0 and `ref_index=[3, 6]` class 1 respectively.

    References
    ----------
    [1] J. Chem. Inf. Comput. Sci. 1997, 37, 6, 1181–1188. https://doi.org/10.1021/ci970282v

    """

    def __init__(
        self,
        r0=None,
        k=10,
        tol=0.01,
        n_iter=10,
        eps=0,
        p=2,
        random_seed=42,
        ref_index=None,
        fun_dist=None,
    ):
        """
        Initialize class.

        Parameters
        ----------
        r0 : float, optional
            Initial guess of radius for OptiSim algorithm. No points within this distance of an
            already selected point can be selected. If `None`, the maximum range of features and
            the size of subset are used to calculate the initial radius. This radius is optimized
            to result in the desired number of samples selected, if possible.
        k : int, optional
            Amount of points to add to subsample before selecting one of the points with the
            greatest minimum distance to the previously selected points.
        tol : float, optional
            Percentage error of number of samples actually selected from number of samples
            requested.
        n_iter : int, optional
            Number of iterations to execute when optimizing the size of exclusion radius.
        p : float, optional
            This is `p` argument of scipy.spatial.KDTree.query_ball_point method denoting
            which Minkowski p-norm to use. Should be in the range [1, inf]. A finite large p may
            cause a ValueError if overflow can occur.
        eps : nonnegative float, optional
            This is `eps` argument of scipy.spatial.KDTree.query_ball_point method denoting
            approximate nearest neighbor search for eliminating close points. Branches of the tree
            are not explored if their nearest points are further than r / (1 + eps), and branches
            are added in bulk if their furthest points are nearer than r * (1 + eps).
        random_seed : int, optional
            Seed for random selection of points be evaluated.
        ref_index: int, list, optional
            Index of the reference sample to start the selection algorithm from.
            It can be an integer, or a list of integers or None. When None, the medoid center is chosen as the reference
            sample.
            When the `ref_index` is a list for multiple classes, it will be shared among all clusters.
            If we want to use different reference indices for each class, we can perform the subset
            selection for each class separately where different `ref_index` parameters can be used.
            For example, if we have two classes, we can pass `ref_index=[0, 1]` to select samples
            from class 0 and `ref_index=[3, 6]` class 1 respectively.
        fun_dist : callable, optional
            Function for calculating the pairwise distance between sample points to be used in
            calculating the medoid. `fun_dist(x) -> x_dist` takes a 2D feature array of shape
            (n_samples, n_features) and returns a 2D distance array of shape (n_samples, n_samples).

        """
        self.r0 = r0
        self.r = r0
        self.ref_index = ref_index
        self.n_iter = n_iter
        self.k = k
        self.tol = tol
        self.eps = eps
        self.p = p
        self.random_seed = random_seed
        self.fun_dist = fun_dist

    def algorithm(self, x, max_size) -> Union[List, Iterable]:
        """Return selected sample indices based on OptiSim algorithm.

        Parameters
        ----------
        x: ndarray of shape (n_samples, n_features)
            Feature matrix of `n_samples` samples in `n_features` dimensional feature space.
        max_size : int
            Maximum number of samples to select.

        Returns
        -------
        selected : Union[List, Iterable]
            List of indices of selected sample indices.

        """
        # set up reference index
        selected = get_initial_selection(x=x, x_dist=None, ref_index=self.ref_index, fun_dist=None)
        count = 1

        # establish a kd-tree for nearest-neighbor lookup
        tree = spatial.KDTree(x)
        # use a random number generator that will be used to randomly select points
        rng = np.random.default_rng(seed=self.random_seed)

        n_samples = len(x)
        # bv will serve as a mask to discard points within radius r of previously selected points
        bv = np.zeros(n_samples)
        candidates = list(range(n_samples))
        # determine which points are within radius r of initial point
        # note: workers=-1 uses all available processors/CPUs
        index_remove = tree.query_ball_point(
            x[self.ref_index], self.r, eps=self.eps, p=self.p, workers=-1
        )
        # exclude points within radius r of initial point from list of candidates using bv mask
        for idx in index_remove:
            bv[idx] = 1
        candidates = np.ma.array(candidates, mask=bv)

        # while there are still remaining candidates to be selected
        # compressed returns all the non-masked data as a 1-D array
        while len(candidates.compressed()) > 0:
            # randomly select samples from list of candidates
            try:
                sublist = rng.choice(candidates.compressed(), size=self.k, replace=False)
            except ValueError:
                sublist = candidates.compressed()

            # create a new kd-tree for nearest neighbor lookup with candidates
            new_tree = spatial.KDTree(x[selected])
            # query the kd-tree for nearest neighbors to selected samples
            # note: workers=-1 uses all available processors/CPUs
            search, _ = new_tree.query(x[sublist], eps=self.eps, p=self.p, workers=-1)
            # identify the nearest neighbor with the largest distance from previously selected samples
            best_idx = sublist[np.argmax(search)]
            selected.append(best_idx)

            count += 1
            if count > max_size:
                # do this if you have reached the maximum number of points selected
                return selected

            # eliminate all samples within radius r of the selected sample
            index_remove = tree.query_ball_point(
                x[best_idx], self.r, eps=self.eps, p=self.p, workers=-1
            )
            for idx in index_remove:
                bv[idx] = 1
            candidates = np.ma.array(candidates, mask=bv)

        return selected

    def select_from_cluster(self, x, size, labels=None) -> Union[List, Iterable]:
        """Return selected samples from a cluster based on OptiSim algorithm.

        Parameters
        ----------
        x: ndarray of shape (n_samples, n_features)
            Feature matrix of `n_samples` samples in `n_features` dimensional feature space.
        size : int
            Number of samples to be selected.
        labels: np.ndarray
            Indices of samples that form a cluster.

        Returns
        -------
        selected : Union[List, Iterable]
            List of indices of selected samples.

        """
        if self.ref_index is not None and self.ref_index >= len(x):
            raise ValueError(
                f"ref_index is not less than the number of samples; {self.ref_index} >= {len(x)}."
            )
        # pass subset of x to optimize_radius if labels is not None
        if labels is not None:
            x = x[labels]
        # reset radius to initial value (this is important when sampling multiple clusters)
        self.r = self.r0
        return optimize_radius(self, x, size, labels)


class DISE(SelectionBase):
    """
    Select samples using Directed Sphere Exclusion (DISE) algorithm.

    In a nutshell, this algorithm iteratively excludes any sample within a given radius from
    any already selected sample. The radius of the exclusion sphere is an adjustable parameter.
    Compared to Sphere Exclusion algorithm, the Directed Sphere Exclusion algorithm achieves a
    more evenly distributed subset selection by abandoning the random selection approach and
    instead imposing a directed selection.

    Reference sample is chosen based on the `ref_index`, which is excluded from the selected
    subset. All samples are sorted (ascending order) based on their Minkowski p-norm distance
    from the reference sample. Looping through sorted samples, the sample is selected if it is
    not already excluded. If selected, all its neighboring samples within a sphere of radius r
    (i.e., exclusion sphere) are excluded from being selected. When the selected number of points
    is greater than specified subset `size`, the selection process terminates. The `r0` is used
    as the initial radius of exclusion sphere, however, it is optimized to select the desired
    number of samples.

    References
    ----------
    Gobbi, A., and Lee, M.-L. (2002). DISE: directed sphere exclusion.
    Journal of Chemical Information and Computer Sciences,
    43(1), 317–323. https://doi.org/10.1021/ci025554v

    """

    def __init__(self, r0=None, ref_index=None, tol=0.05, n_iter=10, p=2.0, eps=0.0, fun_dist=None):
        """
        Initialize class.

        Parameters
        ----------
        r0: float, optional
            Initial guess for radius of the exclusion sphere.
        ref_index: int, list, optional
            Index of the reference sample to start the selection algorithm from.
            It can be an integer, or a list of integers or None. When None, the medoid center is chosen as the reference
            sample.
            When the `ref_index` is a list for multiple classes, it will be shared among all clusters.
            If we want to use different reference indices for each class, we can perform the subset
            selection for each class separately where different `ref_index` parameters can be used.
            For example, if we have two classes, we can pass `ref_index=[0, 1]` to select samples
            from class 0 and `ref_index=[3, 6]` class 1 respectively.
        tol: float, optional
            Percentage tolerance of sample size error. Given a subset size, the selected size
            will be within size * (1 - tol) and size * (1 + tol).
        n_iter: int, optional
            Number of iterations for optimizing the radius of exclusion sphere.
        p: float, optional
            This is `p` argument of scipy.spatial.KDTree.query_ball_point method denoting
            which Minkowski p-norm to use. The values of `p` should be within [1, inf].
            A finite large p may cause a ValueError if overflow can occur. Default is 2.0.
        eps: nonnegative float, optional
            This is `eps` argument of scipy.spatial.KDTree.query_ball_point method denoting
            approximate nearest neighbor search for eliminating close points. Branches of the tree
            are not explored if their nearest points are further than r / (1 + eps), and branches
            are added in bulk if their furthest points are nearer than r * (1 + eps).
        fun_dist: callable, optional
            Function for calculating the distances between sample points. When `fun_dist` is `None`,
            the Minkowski p-norm distance is used. Default is None.

        """
        self.r0 = r0
        self.r = r0
        self.ref_index = ref_index
        self.tol = tol
        self.n_iter = n_iter
        self.p = p
        self.eps = eps

        # if fun_dist is None:
        #     self.fun_dist = spatial.distance.pdist
        # else:
        #     self.fun_dist = fun_dist
        self.fun_dist = fun_dist

    def algorithm(self, x, max_size) -> Union[List, Iterable]:
        """Return selected samples based on directed sphere exclusion algorithm.

        Parameters
        ----------
        x: ndarray of shape (n_samples, n_features)
           Feature matrix of `n_samples` samples in `n_features` dimensional space.
        max_size: int
            Maximum number of samples to select.

        Returns
        -------
        selected: Union[List, Iterable]
            List of indices of selected samples.

        """
        if self.fun_dist is None:
            distances = spatial.distance.squareform(
                spatial.distance.pdist(x, metric="minkowski", p=self.p)
            )
        else:
            distances = self.fun_dist(x)

        # set up the ref_index as when is None
        if self.ref_index is None:
            self.ref_index = get_initial_selection(
                x=None,
                x_dist=distances,
                ref_index=self.ref_index,
                fun_dist=None,
            )
        # set up the ref_index for integer and list of integers
        elif isinstance(self.ref_index, (int, list)):
            self.ref_index = get_initial_selection(
                x=x,
                x_dist=distances,
                ref_index=self.ref_index,
                fun_dist=None,
            )
        # not supported ref_index
        else:
            raise ValueError(
                "The provided reference indices are not supported in the current implementation."
            )

        # # calculate distance of all samples from reference sample; distance is a (n_samples,) array
        # # this includes the distance of reference sample from itself, which is 0
        # distances = spatial.minkowski_distance(x[self.ref_index], x, p=self.p)
        distances_ref = distances[self.ref_index[0], :]

        # get sorted index of samples based on their distance from reference (closest to farthest)
        # the first index will be the ref_index which has distance of zero
        index_sorted = np.argsort(distances_ref)
        assert index_sorted[0] == self.ref_index
        # construct KDTree for quick nearest-neighbor lookup
        kdtree = spatial.KDTree(x)

        # construct bitarray to track selected samples (1 means exclude)
        bv = bitarray.bitarray(list(np.zeros(len(x), dtype=int)))

        # the neighbours of the ref_index are going to be excluded in the first iteration
        # and ref_index is going to be added to the selected list
        selected = []
        for idx in index_sorted:
            # select sample if it is not already excluded from consideration
            # indexing a single item of a bitarray will always return an integer
            if bv[idx] == 0:
                selected.append(idx)
                # return indices of selected samples, if desired number is selected
                if len(selected) > max_size:
                    return selected
                # find index of all samples within radius of sample idx (this includes the sample index itself)
                index_exclude = kdtree.query_ball_point(
                    x[idx], self.r, eps=self.eps, p=self.p, workers=-1
                )
                # exclude samples within radius r of sample idx (measure by Minkowski p-norm) from
                # future consideration by setting their bitarray value to 1
                for index in index_exclude:
                    bv[index] = 1

        return selected

    def select_from_cluster(self, x, size, labels=None) -> Union[List, Iterable]:
        """Return selected samples from a cluster based on directed sphere exclusion algorithm

        Parameters
        ----------
        x: ndarray of shape (n_samples, n_features)
           Feature matrix of `n_samples` samples in `n_features` dimensional space.
        size: int
            Number of samples to be selected.
        labels: np.ndarray, optional
            Indices of samples that form a cluster.

        Returns
        -------
        selected: Union[List, Iterable]
            List of indices of selected samples.

        """
        # pass subset of x to optimize_radius if labels is not None
        if labels is not None:
            x = x[labels]

        if x.shape[0] < size:
            raise RuntimeError(
                f"Number of samples is less than the requested "
                f"sample size: {x.shape[0]} < {size}."
            )
        # reset radius to initial value (this is important when sampling multiple clusters)
        self.r = self.r0
        return optimize_radius(self, x, size, labels)


def get_initial_selection(x=None, x_dist=None, ref_index=None, fun_dist=None) -> List:
    """Set up the reference index for selecting.

    Parameters
    ----------
    x: ndarray of shape (n_samples, n_features), optional
        Feature matrix of `n_samples` samples in `n_features` dimensional feature space.
    x_dist: ndarray of shape (n_samples, n_samples), optional
        Pairwise distance matrix between `n_samples` samples.
    ref_index: int, list, optional
        Index of the reference sample to start the selection algorithm from.
        It can be an integer, or a list of integers or None. When None, the medoid center is chosen as the reference
        sample.
        When the `ref_index` is a list for multiple classes, it will be shared among all clusters.
        If we want to use different reference indices for each class, we can perform the subset
        selection for each class separately where different `ref_index` parameters can be used.
        For example, if we have two classes, we can pass `ref_index=[0, 1]` to select samples
        from class 0 and `ref_index=[3, 6]` class 1 respectively.
    func_dist: callable, optional
        Function for calculating the pairwise distance between sample points to be used in
        calculating the medoid. `func_dist(x) -> x_dist` takes a 2D feature array of shape
        (n_samples, n_features) and returns a 2D distance array of shape (n_samples, n_samples).

    Returns
    -------
    initial_selections: List
        List of indices of the initial selected data points.

    """
    # use the medoid center as the reference sample if ref_index is None
    if ref_index is None:
        if x_dist is None:
            x_dist = fun_dist(x)
        # calculate the medoid center
        initial_selections = [int(np.argmin(np.sum(x_dist, axis=0)))]

    # the length of the distance matrix is the number of samples
    if x_dist is not None:
        len_x = len(x_dist)
    else:
        len_x = len(x)

    # when ref_index is an integer, it cannot be negative or greater than the number of samples
    if isinstance(ref_index, int):
        # check if ref_index is a valid index
        if ref_index < 0 or ref_index >= len_x:
            raise ValueError(
                f"The ref_index must be a non-negative integer less than the "
                f"number of samples, got {ref_index} >= {len_x}."
            )
        initial_selections = [int(ref_index)]

    # when ref_index is a list, just use it
    if isinstance(ref_index, list):
        # all elements of ref_index must be integers or float
        if not all(isinstance(i, (int)) for i in ref_index):
            raise ValueError("All elements of ref_index must be integers.")

        # all the elements of ref_index must be greater than or equal to 0 and less than the
        # number of samples
        if np.any(np.array(ref_index) < 0) or np.any(np.array(ref_index) >= len_x):
            raise ValueError(
                f"All elements of ref_index must be greater than or equal to 0 and less than "
                f"the number of samples, got {ref_index}."
            )

        initial_selections = [int(i) for i in ref_index]

    return initial_selections



# -*- coding: utf-8 -*-
#
# The Selector is a Python library of algorithms for selecting diverse
# subsets of data for machine-learning.
#
# Copyright (C) 2022-2024 The QC-Devs Community
#
# This file is part of Selector.
#
# Selector is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Selector is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Similarity Module."""

from itertools import combinations_with_replacement

import numpy as np

__all__ = [
    "pairwise_similarity_bit",
    "tanimoto",
    "modified_tanimoto",
    "scaled_similarity_matrix",
]


def pairwise_similarity_bit(X: np.array, metric: str) -> np.ndarray:
    """Compute pairwise similarity coefficient matrix.

    Parameters
    ----------
    X : ndarray of shape (n_samples, n_features)
        Feature matrix of `n_samples` samples in `n_features` dimensional space.
    metric : str
        The metric used when calculating similarity coefficients between samples in a feature array.
        Method for calculating similarity coefficient. Options: `"tanimoto"`, `"modified_tanimoto"`.

    Returns
    -------
    s : ndarray of shape (n_samples, n_samples)
        A symmetric similarity matrix between each pair of samples in the feature matrix.
        The diagonal elements are directly computed instead of assuming that they are 1.

    """

    available_methods = {
        "tanimoto": tanimoto,
        "modified_tanimoto": modified_tanimoto,
    }
    if metric not in available_methods:
        raise ValueError(
            f"Argument metric={metric} is not recognized! Choose from {available_methods.keys()}"
        )
    if X.ndim != 2:
        raise ValueError(f"Argument features should be a 2D array, got {X.ndim}")

    # make pairwise m-by-m similarity matrix
    n_samples = len(X)
    s = np.zeros((n_samples, n_samples))
    # compute similarity between all pairs of points (including the diagonal elements)
    for i, j in combinations_with_replacement(range(n_samples), 2):
        s[i, j] = s[j, i] = available_methods[metric](X[i], X[j])
    return s


def tanimoto(a: np.array, b: np.array) -> float:
    r"""Compute Tanimoto coefficient or index (a.k.a. Jaccard similarity coefficient).

    For two binary or non-binary arrays :math:`A` and :math:`B`, Tanimoto coefficient
    is defined as the size of their intersection divided by the size of their union:

    .. math::
        T(A, B) = \frac{| A \cap B|}{| A \cup B |} =
        \frac{| A \cap B|}{|A| + |B| - | A \cap B|} =
        \frac{A \cdot B}{\|A\|^2 + \|B\|^2 - A \cdot B}

    where :math:`A \cdot B = \sum_i{A_i B_i}` and :math:`\|A\|^2 = \sum_i{A_i^2}`.

    Parameters
    ----------
    a : ndarray of shape (n_features,)
        The 1D feature array of sample :math:`A` in an `n_features` dimensional space.
    b : ndarray of shape (n_features,)
        The 1D feature array of sample :math:`B` in an `n_features` dimensional space.

    Returns
    -------
    coeff : float
        Tanimoto coefficient between feature arrays :math:`A` and :math:`B`.

    Bajusz, D., Rácz, A., and Héberger, K.. (2015)
    Why is Tanimoto index an appropriate choice for fingerprint-based similarity calculations?.
    Journal of Cheminformatics 7.

    """
    if a.ndim != 1 or b.ndim != 1:
        raise ValueError(f"Arguments a and b should be 1D arrays, got {a.ndim} and {b.ndim}")
    if a.shape != b.shape:
        raise ValueError(
            f"Arguments a and b should have the same shape, got {a.shape} != {b.shape}"
        )
    coeff = sum(a * b) / (sum(a**2) + sum(b**2) - sum(a * b))
    return coeff


def modified_tanimoto(a: np.array, b: np.array) -> float:
    r"""Compute the modified tanimoto coefficient from bitstring vectors of data points A and B.

    Adjusts calculation of the Tanimoto coefficient to counter its natural bias towards
    shorter vectors using a Bernoulli probability model.

    .. math::
        {mt} = \frac{2-p}{3} T_1 + \frac{1+p}{3} T_0

    where :math:`p` is success probability of independent trials,
    :math:`T_1` is the number of common '1' bits between data points
    (:math:`T_1 = | A \cap B |`), and :math:`T_0` is the number of common '0'
    bits between data points (:math:`T_0 = |(1-A) \cap (1-B)|`).


    Parameters
    ----------
    a : ndarray of shape (n_features,)
        The 1D bitstring feature array of sample :math:`A` in an `n_features` dimensional space.
    b : ndarray of shape (n_features,)
        The 1D bitstring feature array of sample :math:`B` in an `n_features` dimensional space.

    Returns
    -------
    mt : float
        Modified tanimoto coefficient between bitstring feature arrays :math:`A` and :math:`B`.

    Notes
    -----
    The equation above has been derived from

    .. math::
       {mt}_{\alpha} = {\alpha}T_1 + (1-\alpha)T_0

    where :math:`\alpha = \frac{2-p}{3}`. This is done so that the expected value
    of the modified tanimoto, :math:`E(mt)`, remains constant even as the number of
    trials :math:`p` grows larger.

    Fligner, M. A., Verducci, J. S., and Blower, P. E.. (2002)
    A Modification of the Jaccard-Tanimoto Similarity Index for
    Diverse Selection of Chemical Compounds Using Binary Strings.
    Technometrics 44, 110-119.

    """
    if a.ndim != 1:
        raise ValueError(f"Argument `a` should have dimension 1 rather than {a.ndim}.")
    if b.ndim != 1:
        raise ValueError(f"Argument `b` should have dimension 1 rather than {b.ndim}.")
    if a.shape != b.shape:
        raise ValueError(
            f"Arguments a and b should have the same shape, got {a.shape} != {b.shape}"
        )

    n_features = len(a)
    # number of common '1' bits between points A and B
    n_11 = sum(a * b)
    # number of common '0' bits between points A and B
    n_00 = sum((1 - a) * (1 - b))

    # calculate Tanimoto coefficient based on '0' bits
    t_1 = 1
    if n_00 != n_features:
        # bit strings are not all '0's
        t_1 = n_11 / (n_features - n_00)
    # calculate Tanimoto coefficient based on '1' bits
    t_0 = 1
    if n_11 != n_features:
        # bit strings are not all '1's
        t_0 = n_00 / (n_features - n_11)

    # combine into modified tanimoto using Bernoulli Model
    # p = independent success trials
    #       evaluated as total number of '1' bits
    #       divided by 2x the fingerprint length
    p = (n_features - n_00 + n_11) / (2 * n_features)
    # mt = x * T_1 + (1-x) * T_0
    #       x = (2-p)/3 so that E(mt) = 1/3, no matter the value of p
    mt = (((2 - p) / 3) * t_1) + (((1 + p) / 3) * t_0)
    return mt


def scaled_similarity_matrix(X: np.array) -> np.ndarray:
    r"""Compute the scaled similarity matrix.

    .. math::
        X(i,j) = \frac{X(i,j)}{\sqrt{X(i,i)X(j,j)}}

    Parameters
    ----------
    X : ndarray of shape (n_samples, n_samples)
        Similarity matrix of `n_samples`.

    Returns
    -------
    s : ndarray of shape (n_samples, n_samples)
        A scaled symmetric similarity matrix.

    """
    if X.ndim != 2:
        raise ValueError(f"Argument similarity matrix should be a 2D array, got {X.ndim}")

    if X.shape[0] != X.shape[1]:
        raise ValueError(
            f"Argument similarity matrix should be a square matrix (having same number of rows and columns), got {X.shape[0]} and {X.shape[1]}"
        )

    if not (np.all(X >= 0) and np.all(np.diag(X) > 0)):
        raise ValueError(
            "All elements of similarity matrix should be greater than zero and diagonals should be non-zero"
        )

    # scaling does not happen if the matrix is binary similarity matrix with all diagonal elements as 1
    if np.all(np.diag(X) == 1):
        print("No scaling is taking effect")
        return X
    else:
        # make a scaled similarity matrix
        n_samples = len(X)
        s = np.zeros((n_samples, n_samples))
        # calculate the square root of the diagonal elements
        sqrt_diag = np.sqrt(np.diag(X))
        # calculate the product of the square roots of the diagonal elements
        product_sqrt_diag = np.outer(sqrt_diag, sqrt_diag)
        # divide each element of the matrix by the product of the square roots of diagonal elements
        s = X / product_sqrt_diag
        return s


def similarity_index(x: np.array, y: np.array, sim_index: str) -> float:
    """Compute similarity index matrix.

    Parameters
    ----------
    x : ndarray of shape (n_features,)
        Feature array of sample `x` in an `n_features` dimensional space
    y : ndarray of shape (n_features,)
        Feature array of sample `y` in an `n_features` dimensional space
    sim_index : str, optional
        The key with the abbreviation of the similarity index to be used for calculations.
        Possible values are:
            - 'AC': Austin-Colwell
            - 'BUB': Baroni-Urbani-Buser
            - 'CTn': Consoni-Todschini n (n=1,2)
            - 'Fai': Faith
            - 'Gle': Gleason
            - 'Ja': Jaccard
            - 'JT': Jaccard-Tanimoto
            - 'RT': Rogers-Tanimoto
            - 'RR': Russel-Rao
            - 'SM': Sokal-Michener
            - 'SSn': Sokal-Sneath n (n=1,2)
        Default is 'RR'.

    Returns
    -------
    sim : float
        The similarity index value between the feature arrays `x` and `y`.
    """
    # Define the similarity index functions
    similarity_indices = {
        "AC": lambda a, d, dis, p: 2 / np.pi * np.arcsin(((a + d) / p) ** 0.5),
        "BUB": lambda a, d, dis, p: ((a * d) ** 0.5 + a) / ((a * d) ** 0.5 + a + dis),
        "CT1": lambda a, d, dis, p: np.log(1 + a + d) / np.log(1 + p),
        "CT2": lambda a, d, dis, p: (np.log(1 + p) - np.log(1 + dis)) / np.log(1 + p),
        "Fai": lambda a, d, dis, p: (a + 0.5 * d) / p,
        "Gle": lambda a, d, dis, p: 2 * a / (2 * a + dis),
        "Ja": lambda a, d, dis, p: 3 * a / (3 * a + dis),
        "JT": lambda a, d, dis, p: a / (a + dis),
        "RT": lambda a, d, dis, p: (a + d) / (p + dis),
        "RR": lambda a, d, dis, p: a / p,
        "SM": lambda a, d, dis, p: (a + d) / p,
        "SS1": lambda a, d, dis, p: a / (a + 2 * dis),
        "SS2": lambda a, d, dis, p: (2 * (a + d)) / (p + (a + d)),
    }

    if sim_index not in similarity_indices:
        raise ValueError(
            f"Argument sim_index={sim_index} is not recognized! Choose from {similarity_indices.keys()}"
        )
    if x.ndim != 1 or y.ndim != 1:
        raise ValueError(f"Arguments x and y should be 1D arrays, got {x.ndim} and {y.ndim}")
    if x.shape != y.shape:
        raise ValueError(
            f"Arguments x and y should have the same shape, got {x.shape} != {y.shape}"
        )
    a, d, dis, p = _compute_base_descriptors(x, y)
    return similarity_indices[sim_index](a, d, dis, p)


def _compute_base_descriptors(x, y):
    """Compute the base descriptors for the similarity indices.

    Parameters
    ----------
    x : ndarray of shape (n_features,)
        Feature array of sample `x` in an `n_features` dimensional space
    y : ndarray of shape (n_features,)
        Feature array of sample `y` in an `n_features` dimensional space

    Returns
    -------
    tuple(int, int, int, int)
        The number of common on bits, number of common off bits, number of 1-0 mismatches, and the
        length of the fingerprint.
    """
    p = len(x)
    a = np.dot(x, y)
    d = np.dot(1 - x, 1 - y)
    dis = p - a - d
    return a, d, dis, p



# -*- coding: utf-8 -*-
#
# The Selector is a Python library of algorithms for selecting diverse
# subsets of data for machine-learning.
#
# Copyright (C) 2022-2024 The QC-Devs Community
#
# This file is part of Selector.
#
# Selector is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Selector is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --

from selector.methods.distance import *
from selector.methods.partition import *
from selector.methods.similarity import *



# -*- coding: utf-8 -*-
#
# The Selector is a Python library of algorithms for selecting diverse
# subsets of data for machine-learning.
#
# Copyright (C) 2022-2024 The QC-Devs Community
#
# This file is part of Selector.
#
# Selector is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Selector is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Test selector/methods/distance.py."""

import numpy as np
import pytest
from numpy.testing import assert_equal, assert_raises
from scipy.spatial.distance import pdist, squareform
from sklearn.metrics import pairwise_distances

from selector.methods.distance import DISE, MaxMin, MaxSum, OptiSim
from selector.methods.tests.common import (
    generate_synthetic_cluster_data,
    generate_synthetic_data,
    get_data_file_path,
)


def test_maxmin():
    """Testing the MaxMin class."""
    # generate random data points belonging to one cluster - pairwise distance matrix
    _, _, arr_dist = generate_synthetic_data(
        n_samples=100,
        n_features=2,
        n_clusters=1,
        pairwise_dist=True,
        metric="euclidean",
        random_state=42,
    )

    # generate random data points belonging to multiple clusters - class labels and pairwise distance matrix
    _, class_labels_cluster, arr_dist_cluster = generate_synthetic_data(
        n_samples=100,
        n_features=2,
        n_clusters=3,
        pairwise_dist=True,
        metric="euclidean",
        random_state=42,
    )

    # use MaxMin algorithm to select points from clustered data
    collector = MaxMin()
    selected_ids = collector.select(
        arr_dist_cluster,
        size=12,
        labels=class_labels_cluster,
        proportional_selection=False,
    )
    # make sure all the selected indices are the same with expectation
    assert_equal(selected_ids, [41, 34, 94, 85, 51, 50, 66, 78, 21, 64, 29, 83])

    # use MaxMin algorithm to select points from non-clustered data
    collector = MaxMin()
    selected_ids = collector.select(arr_dist, size=12)
    # make sure all the selected indices are the same with expectation
    assert_equal(selected_ids, [85, 57, 41, 25, 9, 62, 29, 65, 81, 61, 60, 97])

    # use MaxMin algorithm to select points from non-clustered data with "medoid" as the reference point
    collector_medoid_1 = MaxMin(ref_index=None)
    selected_ids_medoid_1 = collector_medoid_1.select(arr_dist, size=12)
    # make sure all the selected indices are the same with expectation
    assert_equal(selected_ids_medoid_1, [85, 57, 41, 25, 9, 62, 29, 65, 81, 61, 60, 97])

    # use MaxMin algorithm to select points from non-clustered data with "None" for the reference point
    collector_medoid_2 = MaxMin(ref_index=None)
    selected_ids_medoid_2 = collector_medoid_2.select(arr_dist, size=12)
    # make sure all the selected indices are the same with expectation
    assert_equal(selected_ids_medoid_2, [85, 57, 41, 25, 9, 62, 29, 65, 81, 61, 60, 97])

    # use MaxMin algorithm to select points from non-clustered data with float as the reference point
    collector_float = MaxMin(ref_index=85)
    selected_ids_float = collector_float.select(arr_dist, size=12)
    # make sure all the selected indices are the same with expectation
    assert_equal(selected_ids_float, [85, 57, 41, 25, 9, 62, 29, 65, 81, 61, 60, 97])

    # use MaxMin algorithm to select points from non-clustered data with a predefined list as the reference point
    collector_float = MaxMin(ref_index=[85, 57, 41, 25])
    selected_ids_float = collector_float.select(arr_dist, size=12)
    # make sure all the selected indices are the same with expectation
    assert_equal(selected_ids_float, [85, 57, 41, 25, 9, 62, 29, 65, 81, 61, 60, 97])

    # test failing case when ref_index is not a valid index
    with pytest.raises(ValueError):
        collector_float = MaxMin(ref_index=-3)
        selected_ids_float = collector_float.select(arr_dist, size=12)
    # test failing case when ref_index contains a complex number
    with pytest.raises(ValueError):
        collector_float = MaxMin(ref_index=[1 + 5j, 2, 5])
        _ = collector_float.select(arr_dist, size=12)
    # test failing case when ref_index contains a negative number
    with pytest.raises(ValueError):
        collector_float = MaxMin(ref_index=[-1, 2, 5])
        _ = collector_float.select(arr_dist, size=12)

    # test failing case when the number of labels is not equal to the number of samples
    with pytest.raises(ValueError):
        collector_float = MaxMin(ref_index=85)
        _ = collector_float.select(
            arr_dist, size=12, labels=class_labels_cluster[:90], proportional_selection=False
        )

    # use MaxMin algorithm, this time instantiating with a distance metric
    collector = MaxMin(fun_dist=lambda x: pairwise_distances(x, metric="euclidean"))
    simple_coords = np.array([[0, 0], [2, 0], [0, 2], [2, 2], [-10, -10]])
    # provide coordinates rather than pairwise distance matrix to collector
    selected_ids = collector.select(x=simple_coords, size=3)
    # make sure all the selected indices are the same with expectation
    assert_equal(selected_ids, [0, 4, 3])

    # generating mocked clusters
    np.random.seed(42)
    cluster_one = np.random.normal(0, 1, (3, 2))
    cluster_two = np.random.normal(10, 1, (6, 2))
    cluster_three = np.random.normal(20, 1, (10, 2))
    labels_mocked = np.hstack(
        [[0 for i in range(3)], [1 for i in range(6)], [2 for i in range(10)]]
    )
    mocked_cluster_coords = np.vstack([cluster_one, cluster_two, cluster_three])

    # selecting molecules
    collector = MaxMin(lambda x: pairwise_distances(x, metric="euclidean"))
    selected_mocked = collector.select(
        mocked_cluster_coords,
        size=15,
        labels=labels_mocked,
        proportional_selection=False,
    )
    assert_equal(selected_mocked, [0, 1, 2, 3, 4, 5, 6, 7, 8, 16, 15, 10, 13, 9, 18])


def test_maxmin_proportional_selection():
    """Test MaxMin class with proportional selection."""
    # generate the first cluster with 3 points
    coords, labels, cluster_one, cluster_two, cluster_three = generate_synthetic_cluster_data()
    # instantiate the MaxMin class
    collector = MaxMin(fun_dist=lambda x: pairwise_distances(x, metric="euclidean"), ref_index=0)
    # select 6 points with proportional selection from each cluster
    selected_ids = collector.select(
        coords,
        size=6,
        labels=labels,
        proportional_selection=True,
    )
    # make sure all the selected indices are the same with expectation
    assert_equal(selected_ids, [0, 3, 8, 9, 17, 13])
    # check how many points are selected from each cluster
    assert_equal(len(selected_ids), 6)
    # check the number of points selected from cluster one
    assert_equal((labels[selected_ids] == 0).sum(), 1)
    # check the number of points selected from cluster two
    assert_equal((labels[selected_ids] == 1).sum(), 2)
    # check the number of points selected from cluster three
    assert_equal((labels[selected_ids] == 2).sum(), 3)


def test_maxmin_proportional_selection_imbalance_1():
    """Test MaxMin class with proportional selection with imbalance case 1."""
    # load three-cluster data from file
    # 2 from class 0, 10 from class 1, 40 from class 2
    coords_file_path = get_data_file_path("coords_imbalance_case1.txt")
    coords = np.genfromtxt(coords_file_path, delimiter=",", skip_header=0)
    labels_file_path = get_data_file_path("labels_imbalance_case1.txt")
    labels = np.genfromtxt(labels_file_path, delimiter=",", skip_header=0)

    # instantiate the MaxMin class
    collector = MaxMin(fun_dist=lambda x: pairwise_distances(x, metric="euclidean"), ref_index=0)
    # select 12 points with proportional selection from each cluster
    selected_ids = collector.select(
        coords,
        size=9,
        labels=labels,
        proportional_selection=True,
    )

    # make sure all the selected indices are the same with expectation
    assert_equal(selected_ids, [0, 2, 6, 12, 15, 38, 16, 41, 36])
    # check how many points are selected from each cluster
    assert_equal(len(selected_ids), 9)
    # check the number of points selected from cluster one
    assert_equal((labels[selected_ids] == 0).sum(), 1)
    # check the number of points selected from cluster two
    assert_equal((labels[selected_ids] == 1).sum(), 2)
    # check the number of points selected from cluster three
    assert_equal((labels[selected_ids] == 2).sum(), 6)


def test_maxmin_proportional_selection_imbalance_2():
    """Test MaxMin class with proportional selection with imbalance case 2."""
    # load three-cluster data from file
    # 3 from class 0, 11 from class 1, 40 from class 2
    coords_file_path = get_data_file_path("coords_imbalance_case2.txt")
    coords = np.genfromtxt(coords_file_path, delimiter=",", skip_header=0)
    labels_file_path = get_data_file_path("labels_imbalance_case2.txt")
    labels = np.genfromtxt(labels_file_path, delimiter=",", skip_header=0)

    # instantiate the MaxMin class
    collector = MaxMin(fun_dist=lambda x: pairwise_distances(x, metric="euclidean"), ref_index=0)
    # select 12 points with proportional selection from each cluster
    selected_ids = collector.select(
        coords,
        size=14,
        labels=labels,
        proportional_selection=True,
    )

    # # make sure all the selected indices are the same with expectation
    assert_equal(selected_ids, [0, 3, 9, 6, 14, 36, 53, 17, 44, 23, 28, 50, 52, 49])
    print(f"selected_ids: {selected_ids}")
    # check how many points are selected from each cluster
    assert_equal(len(selected_ids), 14)
    # check the number of points selected from cluster one
    assert_equal((labels[selected_ids] == 0).sum(), 1)
    # check the number of points selected from cluster two
    assert_equal((labels[selected_ids] == 1).sum(), 3)
    # check the number of points selected from cluster three
    assert_equal((labels[selected_ids] == 2).sum(), 10)


def test_maxmin_invalid_input():
    """Testing MaxMin class with invalid input."""
    # case when the distance matrix is not square
    x_dist = np.array([[0, 1], [1, 0], [4, 9]])
    with pytest.raises(ValueError):
        collector = MaxMin(ref_index=0)
        _ = collector.select(x_dist, size=2)

    # case when the distance matrix is not symmetric
    x_dist = np.array([[0, 1, 2, 1], [1, 1, 0, 3], [4, 9, 4, 0], [6, 5, 6, 7]])
    with pytest.raises(ValueError):
        collector = MaxMin(ref_index=0)
        _ = collector.select(x_dist, size=2)


def test_maxsum_clustered_data():
    """Testing MaxSum class."""
    # generate random data points belonging to multiple clusters - coordinates and class labels
    coords_cluster, class_labels_cluster, coords_cluster_dist = generate_synthetic_data(
        n_samples=100,
        n_features=2,
        n_clusters=3,
        pairwise_dist=True,
        metric="euclidean",
        random_state=42,
    )

    # use MaxSum algorithm to select points from clustered data, instantiating with euclidean distance metric
    collector = MaxSum(fun_dist=lambda x: pairwise_distances(x, metric="euclidean"), ref_index=None)
    selected_ids = collector.select(
        coords_cluster,
        size=12,
        labels=class_labels_cluster,
        proportional_selection=False,
    )
    # make sure all the selected indices are the same with expectation
    assert_equal(selected_ids, [41, 34, 85, 94, 51, 50, 78, 66, 21, 64, 0, 83])

    # use MaxSum algorithm to select points from clustered data without instantiating with euclidean distance metric
    collector = MaxSum(ref_index=None)
    selected_ids = collector.select(
        coords_cluster_dist,
        size=12,
        labels=class_labels_cluster,
        proportional_selection=False,
    )
    # make sure all the selected indices are the same with expectation
    assert_equal(selected_ids, [41, 34, 85, 94, 51, 50, 78, 66, 21, 64, 0, 83])

    # check that ValueError is raised when number of points requested is greater than number of points in array
    with pytest.raises(ValueError):
        _ = collector.select_from_cluster(
            coords_cluster,
            size=101,
            labels=class_labels_cluster,
        )


def test_maxsum_non_clustered_data():
    """Testing MaxSum class with non-clustered data."""
    # generate random data points belonging to one cluster - coordinates
    coords, _, _ = generate_synthetic_data(
        n_samples=100,
        n_features=2,
        n_clusters=1,
        pairwise_dist=True,
        metric="euclidean",
        random_state=42,
    )
    # use MaxSum algorithm to select points from non-clustered data, instantiating with euclidean distance metric
    collector = MaxSum(fun_dist=lambda x: pairwise_distances(x, metric="euclidean"), ref_index=None)
    selected_ids = collector.select(coords, size=12)
    # make sure all the selected indices are the same with expectation
    assert_equal(selected_ids, [85, 57, 25, 41, 95, 9, 21, 8, 13, 68, 37, 54])

    # use MaxSum algorithm to select points from non-clustered data, instantiating with euclidean
    # distance metric and using "medoid" as the reference point
    collector = MaxSum(fun_dist=lambda x: pairwise_distances(x, metric="euclidean"), ref_index=None)
    selected_ids = collector.select(coords, size=12)
    # make sure all the selected indices are the same with expectation
    assert_equal(selected_ids, [85, 57, 25, 41, 95, 9, 21, 8, 13, 68, 37, 54])

    # use MaxSum algorithm to select points from non-clustered data, instantiating with euclidean
    # distance metric and using a list as the reference points
    collector = MaxSum(
        fun_dist=lambda x: pairwise_distances(x, metric="euclidean"),
        ref_index=[85, 57, 25, 41, 95],
    )
    selected_ids = collector.select(coords, size=12)
    # make sure all the selected indices are the same with expectation
    assert_equal(selected_ids, [85, 57, 25, 41, 95, 9, 21, 8, 13, 68, 37, 54])

    # use MaxSum algorithm to select points from non-clustered data, instantiating with euclidean
    # distance metric and using an invalid reference point
    with pytest.raises(ValueError):
        collector = MaxSum(
            fun_dist=lambda x: pairwise_distances(x, metric="euclidean"), ref_index=-1
        )
        selected_ids = collector.select(coords, size=12)


def test_maxsum_invalid_input():
    """Testing MaxSum class with invalid input."""
    # case when the distance matrix is not square
    x_dist = np.array([[0, 1], [1, 0], [4, 9]])
    with pytest.raises(ValueError):
        collector = MaxSum(ref_index=0)
        _ = collector.select(x_dist, size=2)

    # case when the distance matrix is not square
    x_dist = np.array([[0, 1, 2], [1, 0, 3], [4, 9, 0], [5, 6, 7]])
    with pytest.raises(ValueError):
        collector = MaxSum(ref_index=0)
        _ = collector.select(x_dist, size=2)

    # case when the distance matrix is not symmetric
    x_dist = np.array([[0, 1, 2, 1], [1, 1, 0, 3], [4, 9, 4, 0], [6, 5, 6, 7]])
    with pytest.raises(ValueError):
        collector = MaxSum(ref_index=0)
        _ = collector.select(x_dist, size=2)


def test_maxsum_proportional_selection():
    """Test MaxSum class with proportional selection."""
    # generate the first cluster with 3 points
    coords, labels, cluster_one, cluster_two, cluster_three = generate_synthetic_cluster_data()
    # instantiate the MaxSum class
    collector = MaxSum(fun_dist=lambda x: pairwise_distances(x, metric="euclidean"), ref_index=0)
    # select 6 points with proportional selection from each cluster
    selected_ids = collector.select(
        coords,
        size=6,
        labels=labels,
        proportional_selection=True,
    )
    # make sure all the selected indices are the same with expectation
    assert_equal(selected_ids, [0, 3, 8, 9, 17, 10])
    # check how many points are selected from each cluster
    assert_equal(len(selected_ids), 6)
    # check the number of points selected from cluster one
    assert_equal((labels[selected_ids] == 0).sum(), 1)
    # check the number of points selected from cluster two
    assert_equal((labels[selected_ids] == 1).sum(), 2)
    # check the number of points selected from cluster three
    assert_equal((labels[selected_ids] == 2).sum(), 3)


def test_optisim():
    """Testing OptiSim class."""
    # generate random data points belonging to one cluster - coordinates and pairwise distance matrix
    coords, _, arr_dist = generate_synthetic_data(
        n_samples=100,
        n_features=2,
        n_clusters=1,
        pairwise_dist=True,
        metric="euclidean",
        random_state=42,
    )

    # generate random data points belonging to multiple clusters - coordinates and class labels
    coords_cluster, class_labels_cluster, _ = generate_synthetic_data(
        n_samples=100,
        n_features=2,
        n_clusters=3,
        pairwise_dist=True,
        metric="euclidean",
        random_state=42,
    )

    # use OptiSim algorithm to select points from clustered data
    collector = OptiSim(ref_index=0)
    selected_ids = collector.select(coords_cluster, size=12, labels=class_labels_cluster)
    # make sure all the selected indices are the same with expectation
    # assert_equal(selected_ids, [2, 85, 86, 59, 1, 66, 50, 68, 0, 64, 83, 72])

    # use OptiSim algorithm to select points from non-clustered data
    collector = OptiSim(ref_index=0)
    selected_ids = collector.select(coords, size=12)
    # make sure all the selected indices are the same with expectation
    assert_equal(selected_ids, [0, 8, 55, 37, 41, 13, 12, 42, 6, 30, 57, 76])

    # check if OptiSim gives same results as MaxMin for k=>infinity
    collector = OptiSim(ref_index=85, k=999999)
    selected_ids_optisim = collector.select(coords, size=12)
    collector = MaxMin()
    selected_ids_maxmin = collector.select(arr_dist, size=12)
    assert_equal(selected_ids_optisim, selected_ids_maxmin)

    # test with invalid ref_index
    with pytest.raises(ValueError):
        collector = OptiSim(ref_index=10000)
        _ = collector.select(coords, size=12)


def test_optisim_proportional_selection():
    """Test OptiSim class with proportional selection."""
    # generate the first cluster with 3 points
    coords, labels, cluster_one, cluster_two, cluster_three = generate_synthetic_cluster_data()
    # instantiate the Optisim class
    collector = OptiSim(fun_dist=lambda x: pairwise_distances(x, metric="euclidean"), ref_index=0)
    # select 6 points with proportional selection from each cluster
    selected_ids = collector.select(
        coords,
        size=6,
        labels=labels,
        proportional_selection=True,
    )
    # make sure all the selected indices are the same with expectation
    assert_equal(selected_ids, [0, 3, 8, 9, 17, 13])
    # check how many points are selected from each cluster
    assert_equal(len(selected_ids), 6)
    # check the number of points selected from cluster one
    assert_equal((labels[selected_ids] == 0).sum(), 1)
    # check the number of points selected from cluster two
    assert_equal((labels[selected_ids] == 1).sum(), 2)
    # check the number of points selected from cluster three
    assert_equal((labels[selected_ids] == 2).sum(), 3)


def test_directed_sphere_size_error():
    """Test DirectedSphereExclusion error when too many points requested."""
    x = np.array([[1, 9]] * 100)
    collector = DISE()
    assert_raises(ValueError, collector.select, x, size=105)


def test_directed_sphere_same_number_of_pts():
    """Test DirectSphereExclusion with `size` = number of points in dataset."""
    # (0,0) as the reference point
    x = np.array([[0, 0], [0, 1], [0, 2], [0, 3]])
    collector = DISE(r0=1, tol=0, ref_index=0)
    selected = collector.select(x, size=2)
    assert_equal(selected, [0, 2])
    assert_equal(collector.r, 1)


def test_directed_sphere_same_number_of_pts_None():
    """Test DirectSphereExclusion with `size` = number of points in dataset with the ref_index None."""
    # None as the reference point
    x = np.array([[0, 0], [0, 1], [0, 2], [0, 3], [0, 4]])
    collector = DISE(r0=1, tol=0, ref_index=None)
    selected = collector.select(x, size=3)
    assert_equal(selected, [2, 0, 4])
    assert_equal(collector.r, 1)


def test_directed_sphere_exclusion_select_more_number_of_pts():
    """Test DirectSphereExclusion on points on the line with `size` < number of points in dataset."""
    # (0,0) as the reference point
    x = np.array([[0, 0], [0, 1], [0, 2], [0, 3], [0, 4], [0, 5], [0, 6]])
    collector = DISE(r0=0.5, tol=0, ref_index=0)
    selected = collector.select(x, size=3)
    expected = [0, 3, 6]
    assert_equal(selected, expected)
    assert_equal(collector.r, 2.0)


def test_directed_sphere_exclusion_on_line_with_smaller_radius():
    """Test Direct Sphere Exclusion on points on line with smaller distribution than the radius."""
    # (0,0) as the reference point
    x = np.array(
        [
            [0, 0],
            [0, 1],
            [0, 1.1],
            [0, 1.2],
            [0, 2],
            [0, 3],
            [0, 3.1],
            [0, 3.2],
            [0, 4],
            [0, 5],
            [0, 6],
        ]
    )
    collector = DISE(r0=0.5, tol=1, ref_index=1)
    selected = collector.select(x, size=3)
    expected = [1, 5, 9]
    assert_equal(selected, expected)
    assert_equal(collector.r, 1.0)


def test_directed_sphere_on_line_with_larger_radius():
    """Test Direct Sphere Exclusion on points on the line with a too large radius size."""
    # (0,0) as the reference point
    x = np.array(
        [
            [0, 0],
            [0, 1],
            [0, 1.1],
            [0, 1.2],
            [0, 2],
            [0, 3],
            [0, 3.1],
            [0, 3.2],
            [0, 4],
            [0, 5],
        ]
    )
    collector = DISE(r0=2.0, tol=0, p=2.0, ref_index=1)
    selected = collector.select(x, size=3)
    expected = [1, 5, 9]
    assert_equal(selected, expected)
    assert_equal(collector.r, 1.0)


def test_directed_sphere_dist_func():
    """Test Direct Sphere Exclusion with a distance function."""
    # (0,0) as the reference point
    x = np.array([[0, 0], [0, 1], [0, 2], [0, 3], [0, 4], [0, 5], [0, 6]])
    collector = DISE(
        r0=0.5,
        tol=0,
        ref_index=0,
        fun_dist=lambda x: squareform(pdist(x, metric="minkowski", p=0.1)),
    )
    selected = collector.select(x, size=3)
    expected = [0, 3, 6]
    assert_equal(selected, expected)
    assert_equal(collector.r, 2.0)


def test_directed_sphere_proportional_selection():
    """Test DISE class with proportional selection."""
    # generate the first cluster with 3 points
    coords, labels, cluster_one, cluster_two, cluster_three = generate_synthetic_cluster_data()
    # instantiate the DISE class
    collector = DISE(fun_dist=lambda x: pairwise_distances(x, metric="euclidean"), ref_index=0)
    # select 6 points with proportional selection from each cluster
    selected_ids = collector.select(
        coords,
        size=6,
        labels=labels,
        proportional_selection=True,
    )
    # make sure all the selected indices are the same with expectation
    assert_equal(selected_ids, [0, 3, 7, 9, 12, 15])
    # check how many points are selected from each cluster
    assert_equal(len(selected_ids), 6)
    # check the number of points selected from cluster one
    assert_equal((labels[selected_ids] == 0).sum(), 1)
    # check the number of points selected from cluster two
    assert_equal((labels[selected_ids] == 1).sum(), 2)
    # check the number of points selected from cluster three
    assert_equal((labels[selected_ids] == 2).sum(), 3)



# -*- coding: utf-8 -*-
#
# The Selector is a Python library of algorithms for selecting diverse
# subsets of data for machine-learning.
#
# Copyright (C) 2022-2024 The QC-Devs Community
#
# This file is part of Selector.
#
# Selector is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Selector is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Module for converting similarity measures to distance/dissimilarity measures."""

from typing import Union

import numpy as np

__all__ = [
    "sim_to_dist",
    "reverse",
    "reciprocal",
    "exponential",
    "gaussian",
    "correlation",
    "transition",
    "co_occurrence",
    "gravity",
    "probability",
    "covariance",
]


def sim_to_dist(
    x: Union[int, float, np.ndarray], metric: str, scaling_factor: float = 1.0
) -> Union[float, np.ndarray]:
    """Convert similarity coefficients to distance array.

    Parameters
    ----------
    x : float or ndarray
        A similarity value as float, or a 1D or 2D array of similarity values.
        If 2D, the array is assumed to be symmetric.
    metric : str
        String or integer specifying which conversion metric to use.
        Supported metrics are "reverse", "reciprocal", "exponential",
        "gaussian", "membership", "correlation", "transition", "co-occurrence",
        "gravity", "confusion", "probability", and "covariance".
    scaling_factor : float, optional
        Scaling factor for the distance array. Default is 1.0.

    Returns
    -------
    dist : float or ndarray
         Distance value or array.
    """
    # scale the distance matrix
    x = x * scaling_factor

    frequency = {
        "transition": transition,
        "co-occurrence": co_occurrence,
        "gravity": gravity,
    }
    method_dict = {
        "reverse": reverse,
        "reciprocal": reciprocal,
        "exponential": exponential,
        "gaussian": gaussian,
        "correlation": correlation,
        "probability": probability,
        "covariance": covariance,
    }

    # check if x is a single value
    single_value = False
    if isinstance(x, (float, int)):
        x = np.array([[x]])
        single_value = True

    # check x
    if not isinstance(x, np.ndarray):
        raise ValueError(f"Argument x should be a numpy array instead of {type(x)}")
    # check that x is a valid array
    if x.ndim != 1 and x.ndim != 2:
        raise ValueError(f"Argument x should either have 1 or 2 dimensions, got {x.ndim}.")
    if x.ndim == 1 and metric in ["co-occurrence", "gravity"]:
        raise ValueError(f"Argument x should be a 2D array when using the {metric} metric.")
    # check if x is symmetric
    if x.ndim == 2 and not np.allclose(x, x.T):
        raise ValueError("Argument x should be a symmetric array.")

    # call correct metric function
    if metric in frequency:
        if np.any(x <= 0):
            raise ValueError(
                "There is a negative or zero value in the input. Please "
                "make sure all frequency values are positive."
            )
        dist = frequency[metric](x)
    elif metric in method_dict:
        dist = method_dict[metric](x)
    elif metric == "membership" or metric == "confusion":
        if np.any(x < 0) or np.any(x > 1):
            raise ValueError(
                "There is an out of bounds value. Please make "
                "sure all input values are between [0, 1]."
            )
        dist = 1 - x
    # unsupported metric
    else:
        raise ValueError(f"{metric} is an unsupported metric.")

    # convert back to float if input was single value
    if single_value:
        dist = dist.item((0, 0))

    return dist


def reverse(x: np.ndarray) -> np.ndarray:
    r"""Calculate distance array from similarity using the reverse method.

    .. math::
        \delta_{ij} = min(s_{ij}) + max(s_{ij}) - s_{ij}

    where :math:`\delta_{ij}` is the distance between points :math:`i`
    and :math:`j`, :math:`s_{ij}` is their similarity coefficient,
    and :math:`max` and :math:`min` are the maximum and minimum
    values across the entire similarity array.

    Parameters
    -----------
    x : ndarray
        Similarity array.

    Returns
    -------
    dist : ndarray
        Distance array.

    """
    dist = np.max(x) + np.min(x) - x
    return dist


def reciprocal(x: np.ndarray) -> np.ndarray:
    r"""Calculate distance array from similarity using the reciprocal method.

    .. math::
        \delta_{ij} = \frac{1}{s_{ij}}

    where :math:`\delta_{ij}` is the distance between points :math:`i`
    and :math:`j`, and :math:`s_{ij}` is their similarity coefficient.

    Parameters
    -----------
    x : ndarray
        Similarity array.

    Returns
    -------
    dist : ndarray
        Distance array.
    """

    if np.any(x <= 0):
        raise ValueError(
            "There is an out of bounds value. Please make " "sure all similarities are positive."
        )
    return 1 / x


def exponential(x: np.ndarray) -> np.ndarray:
    r"""Calculate distance matrix from similarity using the exponential method.

    .. math::
        \delta_{ij} = -\ln{\frac{s_{ij}}{max(s_{ij})}}

    where :math:`\delta_{ij}` is the distance between points :math:`i`
    and :math:`j`, and :math:`s_{ij}` is their similarity coefficient.

    Parameters
    -----------
    x : ndarray
        Similarity array.

    Returns
    -------
    dist : ndarray
        Distance array.

    """
    max_sim = np.max(x)
    if max_sim == 0:
        raise ValueError("Maximum similarity in `x` is 0. Distance cannot be computed.")
    dist = -np.log(x / max_sim)
    return dist


def gaussian(x: np.ndarray) -> np.ndarray:
    r"""Calculate distance matrix from similarity using the Gaussian method.

    .. math::
        \delta_{ij} = \sqrt{-\ln{\frac{s_{ij}}{max(s_{ij})}}}

    where :math:`\delta_{ij}` is the distance between points :math:`i`
    and :math:`j`, and :math:`s_{ij}` is their similarity coefficient.

    Parameters
    -----------
    x : ndarray
        Similarity array.

    Returns
    -------
    dist : ndarray
        Distance array.

    """
    max_sim = np.max(x)
    if max_sim == 0:
        raise ValueError("Maximum similarity in `x` is 0. Distance cannot be computed.")
    y = x / max_sim
    dist = np.sqrt(-np.log(y))
    return dist


def correlation(x: np.ndarray) -> np.ndarray:
    r"""Calculate distance array from correlation array.

    .. math::
        \delta_{ij} = \sqrt{1 - r_{ij}}

    where :math:`\delta_{ij}` is the distance between points :math:`i`
    and :math:`j`, and :math:`r_{ij}` is their correlation.

    Parameters
    -----------
    x : ndarray
        Correlation array.

    Returns
    -------
    dist : ndarray
        Distance array.

    """
    if np.any(x < -1) or np.any(x > 1):
        raise ValueError(
            "There is an out of bounds value. Please make "
            "sure all correlations are between [-1, 1]."
        )
    dist = np.sqrt(1 - x)
    return dist


def transition(x: np.ndarray) -> np.ndarray:
    r"""Calculate distance array from frequency using the transition method.

    .. math::
        \delta_{ij} = \frac{1}{\sqrt{f_{ij}}}

    where :math:`\delta_{ij}` is the distance between points :math:`i`
    and :math:`j`, and :math:`f_{ij}` is their frequency.

    Parameters
    -----------
    x : ndarray
        Symmetric frequency array.

    Returns
    -------
    dist : ndarray
        Distance array.

    """
    dist = 1 / np.sqrt(x)
    return dist


def co_occurrence(x: np.ndarray) -> np.ndarray:
    r"""Calculate distance array from frequency using the co-occurrence method.

    .. math::
        \delta_{ij} =  \left(1 + \frac{f_{ij}\sum_{i,j}{f_{ij}}}{\sum_{i}{f_{ij}}\sum_{j}{f_{ij}}} \right)^{-1}

    where :math:`\delta_{ij}` is the distance between points :math:`i`
    and :math:`j`, and :math:`f_{ij}` is their frequency.

    Parameters
    -----------
    x : ndarray
        Frequency array.

    Returns
    -------
    dist : ndarray
        Co-occurrence array.

    """
    # compute sums along each axis
    i = np.sum(x, axis=0, keepdims=True)
    j = np.sum(x, axis=1, keepdims=True)
    # multiply sums to scalar value
    bottom = np.dot(i, j)
    # multiply each element by the sum of entire array
    top = x * np.sum(x)
    # evaluate function as a whole
    dist = (1 + (top / bottom)) ** -1
    return dist


def gravity(x: np.ndarray) -> np.ndarray:
    r"""Calculate distance array from frequency using the gravity method.

    .. math::
        \delta_{ij} = \sqrt{\frac{\sum_{i}{f_{ij}}\sum_{j}{f_{ij}}}
        {f_{ij}\sum_{i,j}{f_{ij}}}}

    where :math:`\delta_{ij}` is the distance between points :math:`i`
    and :math:`j`, and :math:`f_{ij}` is their frequency.

    Parameters
    -----------
    x : ndarray
        Symmetric frequency array.

    Returns
    -------
    dist : ndarray
        Symmetric gravity array.

    """
    # compute sums along each axis
    i = np.sum(x, axis=0, keepdims=True)
    j = np.sum(x, axis=1, keepdims=True)
    # multiply sums to scalar value
    top = np.dot(i, j)
    # multiply each element by the sum of entire array
    bottom = x * np.sum(x)
    # take square root of the fraction
    dist = np.sqrt(top / bottom)
    return dist


def probability(x: np.ndarray) -> np.ndarray:
    r"""Calculate distance array from probability array.

    .. math::
        \delta_{ij} = \sqrt{-\ln{\frac{s_{ij}}{max(s_{ij})}}}

    where :math:`\delta_{ij}` is the distance between points :math:`i`
    and :math:`j`, and :math:`p_{ij}` is their probablity.

    Parameters
    -----------
    x : ndarray
        Symmetric probability array.

    Returns
    -------
    dist : ndarray
        Distance array.

    """
    if np.any(x <= 0) or np.any(x > 1):
        raise ValueError(
            "There is an out of bounds value. Please make "
            "sure all probabilities are between (0, 1]."
        )
    y = np.arcsin(x)
    dist = 1 / np.sqrt(y)
    return dist


def covariance(x: np.ndarray) -> np.ndarray:
    r"""Calculate distance array from similarity using the covariance method.

    .. math::
        \delta_{ij} = \sqrt{s_{ii}+s_{jj}-2s_{ij}}

    where :math:`\delta_{ij}` is the distance between points :math:`i`
    and :math:`j`, :math:`s_{ii}` and :math:`s_{jj}` are the variances
    of feature :math:`i` and feature :math:`j`, and :math:`s_{ij}`
    is the covariance between the two features.

    Parameters
    -----------
    x : ndarray
        Covariance array.

    Returns
    -------
    dist : ndarray
        Distance array.

    """
    variance = np.diag(x).reshape([x.shape[0], 1]) * np.ones([1, x.shape[0]])
    if np.any(variance < 0):
        raise ValueError("Variance of a single variable cannot be negative.")

    dist = variance + variance.T - 2 * x
    dist = np.sqrt(dist)
    return dist



# -*- coding: utf-8 -*-
#
# The Selector is a Python library of algorithms for selecting diverse
# subsets of data for machine-learning.
#
# Copyright (C) 2022-2024 The QC-Devs Community
#
# This file is part of Selector.
#
# Selector is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Selector is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Module for Selection Utilities."""

import warnings

import numpy as np

__all__ = [
    "optimize_radius",
]


def optimize_radius(obj, X, size, cluster_ids=None):
    """Algorithm that uses sphere exclusion for selecting points from cluster.

    Iteratively searches for the optimal radius to obtain the correct number
    of selected samples. If the radius cannot converge to return `size` points,
    the function returns the closest number of samples to `size` as possible.

    Parameters
    ----------
    obj: object
        Instance of `DirectedSphereExclusion` or `OptiSim` selection class.
    X: ndarray of shape (n_samples, n_features)
        Feature matrix of `n_samples` samples in `n_features` dimensional space.
    size: int
        Number of sample points to select (i.e. size of the subset).
    cluster_ids: np.ndarray
        Indices of points that form a cluster.

    Returns
    -------
    selected: list
        List of indices of selected samples.
    """
    if X.shape[0] < size:
        raise RuntimeError(
            f"Size of samples to be selected is greater than existing the number of samples; "
            f"{size} > {X.shape[0]}."
        )
    # set the limits on # of selected points according to the tolerance percentage
    error = size * obj.tol
    lower_size = round(size - error)
    upper_size = round(size + error)

    # select `upper_size` number of samples
    if obj.r is not None:
        # use initial sphere radius
        selected = obj.algorithm(X, upper_size)
    else:
        # calculate a sphere radius based on maximum of n_features range
        # np.ptp returns range of values (maximum - minimum) along an axis
        obj.r = max(np.ptp(X, axis=0)) / size * 3
        selected = obj.algorithm(X, upper_size)

    # return selected if the correct number of samples chosen
    if len(selected) == size:
        return selected

    # optimize radius to select the correct number of samples
    # first, set a sensible range for optimizing r value within that range
    if len(selected) > size:
        # radius should become bigger, b/c too many samples were selected
        bounds = [obj.r, np.inf]
    else:
        # radius should become smaller, b/c too few samples were selected
        bounds = [0, obj.r]

    n_iter = 0
    while (len(selected) < lower_size or len(selected) > upper_size) and n_iter < obj.n_iter:
        # change sphere radius based on the defined bound
        if bounds[1] == np.inf:
            # make sphere radius larger by a factor of 2
            obj.r = bounds[0] * 2
        else:
            # make sphere radius smaller by a factor of 1/2
            obj.r = (bounds[0] + bounds[1]) / 2

        # re-select samples with the new radius
        selected = obj.algorithm(X, upper_size)

        # adjust lower/upper bounds of radius range
        if len(selected) > size:
            bounds[0] = obj.r
        else:
            bounds[1] = obj.r
        n_iter += 1

    # cannot find radius that produces desired number of selected points
    if n_iter >= obj.n_iter and len(selected) != size:
        warnings.warn(
            f"Optimal radius finder failed to converge, selected {len(selected)} points instead "
            f"of requested {size}."
        )

    return selected



# -*- coding: utf-8 -*-
#
# The Selector is a Python library of algorithms for selecting diverse
# subsets of data for machine-learning.
#
# Copyright (C) 2022-2024 The QC-Devs Community
#
# This file is part of Selector.
#
# Selector is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Selector is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Base class for diversity based subset selection."""

import warnings
from abc import ABC, abstractmethod
from typing import List, Iterable, Union

import numpy as np

__all__ = ["SelectionBase"]


class SelectionBase(ABC):
    """Base class for selecting subset of sample points."""

    def select(
        self,
        x: np.ndarray,
        size: int,
        labels: np.ndarray = None,
        proportional_selection: bool = True,
    ) -> Union[List, Iterable]:
        """Return indices representing subset of sample points.

        Parameters
        ----------
        x: ndarray of shape (n_samples, n_features) or (n_samples, n_samples)
            Feature matrix of `n_samples` samples in `n_features` dimensional feature space.
            If fun_distance is `None`, this x is treated as a square pairwise distance matrix.
        size: int
            Number of sample points to select (i.e. size of the subset).
        labels: np.ndarray, optional
            Array of integers or strings representing the labels of the clusters that
            each sample belongs to. If `None`, the samples are treated as one cluster.
            If labels are provided, selection is made from each cluster.
        proportional_selection: bool, optional
            If True, the number of samples to be selected from each cluster is proportional.
            Otherwise, the number of samples to be selected from each cluster is equal.
            Default is True.

        Returns
        -------
        selected: list
            Indices of the selected sample points.
        """
        # check size
        if size > len(x):
            raise ValueError(
                f"Size of subset {size} cannot be larger than number of samples {len(x)}."
            )

        # if labels are not provided, indices selected from one cluster is returned
        if labels is None:
            return self.select_from_cluster(x, size)

        # check labels are consistent with number of samples
        if len(labels) != len(x):
            raise ValueError(
                f"Number of labels {len(labels)} does not match number of samples {len(x)}."
            )

        selected_ids = []

        # compute the number of samples (i.e. population or pop) in each cluster
        unique_labels, unique_label_counts = np.unique(labels, return_counts=True)
        num_clusters = len(unique_labels)
        pop_clusters = dict(zip(unique_labels, unique_label_counts))
        # compute number of samples to be selected from each cluster
        if proportional_selection:
            # make sure that tht total number of samples selected is equal to size
            size_each_cluster = size * unique_label_counts / len(labels)
            # using np.round to get to the nearest integer
            # not using int function directly to avoid truncation of decimal values
            size_each_cluster = np.round(size_each_cluster).astype(int)
            # make sure each cluster has at least one sample
            size_each_cluster[size_each_cluster < 1] = 1

            # the total number of samples selected from all clusters at this point
            size_each_cluster_total = np.sum(size_each_cluster)
            # when the total of data points in each class is less than the required number
            # add one sample to the smallest cluster iteratively until the total is equal to the
            # required number
            if size_each_cluster_total < size:
                while size_each_cluster_total < size:
                    # the number of remaining data points in each cluster
                    size_each_cluster_remaining = unique_label_counts - size_each_cluster_total
                    # skip the clusters with no data points left
                    size_each_cluster_remaining[size_each_cluster_remaining == 0] = np.inf
                    smallest_cluster_index = np.argmin(size_each_cluster_remaining)
                    size_each_cluster[smallest_cluster_index] += 1
                    size_each_cluster_total += 1
            # when the total of data points in each class is more than the required number
            # we need to remove samples from the largest clusters
            elif size_each_cluster_total > size:
                while size_each_cluster_total > size:
                    largest_cluster_index = np.argmax(size_each_cluster)
                    size_each_cluster[largest_cluster_index] -= 1
                    size_each_cluster_total -= 1
            # perfect case where the total is equal to the required number
            else:
                pass
        else:
            size_each_cluster = size // num_clusters

            # update number of samples to select from each cluster based on the cluster population.
            # this is needed when some clusters do not have enough samples in them
            # (pop < size_each_cluster) and needs to be done iteratively until all remaining clusters
            # have at least size_each_cluster samples
            while np.any(
                [value <= size_each_cluster for value in pop_clusters.values() if value != 0]
            ):
                for unique_label in unique_labels:
                    if pop_clusters[unique_label] != 0:
                        # get index of sample labelled with unique_label
                        cluster_ids = np.where(labels == unique_label)[0]
                        if len(cluster_ids) <= size_each_cluster:
                            # all samples in the cluster are selected & population becomes zero
                            selected_ids.append(cluster_ids)
                            pop_clusters[unique_label] = 0
                # update number of samples to be selected from each cluster
                totally_used_clusters = list(pop_clusters.values()).count(0)
                size_each_cluster = (size - len(np.hstack(selected_ids))) // (
                    num_clusters - totally_used_clusters
                )

                warnings.warn(
                    f"Number of molecules in one cluster is less than"
                    f" {size}/{num_clusters}.\nNumber of selected "
                    f"molecules might be less than desired.\nIn order to avoid this "
                    f"problem. Try to use less number of clusters."
                )
            # save the number of samples to be selected from each cluster in an array
            size_each_cluster = np.full(num_clusters, size_each_cluster)

        for unique_label, size_sub in zip(unique_labels, size_each_cluster):
            if pop_clusters[unique_label] != 0:
                # sample size_each_cluster ids from cluster labeled unique_label
                cluster_ids = np.where(labels == unique_label)[0]
                selected = self.select_from_cluster(x, size_sub, cluster_ids)
                selected_ids.append(cluster_ids[selected])

        return np.hstack(selected_ids).flatten().tolist()

    @abstractmethod
    def select_from_cluster(
        self, x: np.ndarray, size: int, labels: np.ndarray = None
    ) -> np.ndarray: # pragma: no cover
        """Return indices representing subset of sample points from one cluster.

        Parameters
        ----------
        x: ndarray of shape (n_samples, n_features) or (n_samples, n_samples)
            Feature matrix of `n_samples` samples in `n_features` dimensional feature space.
            If fun_distance is `None`, this x is treated as a square pairwise distance matrix.
        size: int
            Number of sample points to select (i.e. size of the subset).
        labels: np.ndarray, optional
            Array of integers or strings representing the labels of the clusters that
            each sample belongs to. If `None`, the samples are treated as one cluster.
            If labels are provided, selection is made from each cluster.

        Returns
        -------
        selected: list
            Indices of the selected sample points.
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
#
# The Selector is a Python library of algorithms for selecting diverse
# subsets of data for machine-learning.
#
# Copyright (C) 2022-2024 The QC-Devs Community
#
# This file is part of Selector.
#
# Selector is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Selector is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Module for Partition-Based Selection Methods."""

import collections
import math

import bitarray
import numpy as np
import scipy

from selector.measures.diversity import compute_diversity
from selector.methods.base import SelectionBase

__all__ = [
    "GridPartition",
    "Medoid",
]


class GridPartition(SelectionBase):
    r"""Select subset of sample points using the grid partitioning algorithms.

    Given the number of bins along each axis, samples are partitioned using various methods [1]_:

    - The `equisized_independent` partitions the feature space into bins of equal size along
      each dimension.

    - The `equisized_dependent` partitions the space where the bins can have different length
      in each dimension. I.e., the `l-`th dimension bins depend on the previous dimensions.
      So, the order of features affects the outcome.

    - The `equifrequent_independent` divides the space into bins with approximately equal
      number of sample points in each bin.

    - The `equifrequent_dependent` is similar to `equisized_dependent` where the partition in
      each dimension will depend on the previous dimensions.

    References
    ----------
    .. [1] Bayley, Martin J., and Peter Willett. "Binning schemes for partition-based
           compound selection." Journal of Molecular Graphics and Modelling 17.1 (1999): 10-18.
    """

    def __init__(
        self, nbins_axis: int, bin_method: str = "equisized_independent", random_seed: int = 42
    ):
        """Initialize class.

        Parameters
        ----------
        nbins_axis: int
            Number of bins to partition each axis into. The total number of resulting bins is
            `numb_bins_axis` raised to the power of the dimensionality of the feature space.
        bin_method: str, optional
            Method used to partition the sample points into bins. Options include:
            "equisized_independent", "equisized_dependent", "equifrequent_independent" and
            "equifrequent_dependent".
        random_seed: int, optional
            Seed for random selection of sample points from each bin.
        """
        if not isinstance(nbins_axis, int):
            raise TypeError(f"Number of bins should be integer, got {type(nbins_axis)}.")
        if not isinstance(random_seed, int):
            raise TypeError(f"The random seed should be integer, got {type(random_seed)}.")
        if not isinstance(bin_method, str):
            raise TypeError(f"The bin_method should be a string, got {type(bin_method)}.")
        self.random_seed = random_seed
        self.nbins_axis = nbins_axis
        self.bin_method = bin_method

    @staticmethod
    def partition_points_to_bins_equisized(X, nbins_axis):
        r"""
        Find all bins ids that has points in them and assign each point to each of those bins.

        For each `n_features` dimensions, get the minimum and maximum of feature to and use
        `nbins_axis` to compute length of the bin. Then assign sample points to bins along
        each axis/dimension of feature space.

        Parameters
        ----------
        X: ndarray of shape (n_samples, n_features)
            Feature matrix of `n_samples` samples in `n_features` dimensional space.
        nbins_axis: int
            Number of bins along each axis/dimension of feature space.

        Returns
        -------
        unique_bin_indices: ndarray(int,)
            Unique (without duplication) bin indices that have at least one sample point.
            These are integer tuples :math:`(i_1, \cdot, i_\text{n_features})` with elements
            corresponding to the bin index along each axis/dimension of feature space.
            `inverse_ids` contains indices of `unique_bins_ids` for each of the :math:`N` points that
            it is assigned to.
        inverse_indices: ndarray(int,)
            Indices of the unique bins (along specified axis) that can be used to reconstruct bin
            index of each sample (`unique_bin_indices[inverse_indices]` gives bin index array).
        """
        # find the minimum and maximum of features along axis/dimension
        axis_minimum = np.min(X, axis=0)
        axis_maximum = np.max(X, axis=0)
        bin_length = (axis_maximum - axis_minimum) / nbins_axis
        # assign each sample to a bin along each dimension (floor_divide returns array of integers)
        bin_index = np.floor_divide(X - axis_minimum, bin_length)
        # get unique bin indices (occupied by samples) and indices of the unique array
        # (along specified axis) that can be used to reconstruct bin_index
        # in other words, unique_bin_index[inverse_index] gives back bin_index array
        unique_bin_index, inverse_index = np.unique(bin_index, return_inverse=True, axis=0)
        return unique_bin_index, inverse_index

    @staticmethod
    def partition_points_to_bins_equifrequent(X, nbins_axis):
        r"""
        Find all bins ids that contains points using the equifrequent method.

        The equifrequent method partitions each bin to have equal number of points.
        This is done by doing a linear interpolation from integer indices and points, where
        it is then evaluated on a uniform grid with number of bins as the spacing in each axis.

        Parameters
        ----------
        X: ndarray of shape (n_samples, n_features)
           Feature matrix of `n_samples` samples in `n_features` dimensional space.
        nbins_axis: int
            Number of bins along each axis or feature dimension.

        Returns
        -------
        unique_bin_indices: ndarray(int,)
            Unique (without duplication) bin indices that have at least one sample point.
            These are integer tuples :math:`(i_1, \cdot, i_\text{n_features})` with elements
            corresponding to the bin index along each axis/dimension of feature space.
            `inverse_ids` contains indices of `unique_bins_ids` for each of the :math:`N` points that
            it is assigned to.
        inverse_indices: ndarray(int,)
            Indices of the unique bins (along specified axis) that can be used to reconstruct bin
            index of each sample (`unique_bin_indices[inverse_indices]` gives bin index array).
        """
        n_samples = len(X)
        # to obtain the lower and upper range of bins so that each bin has equal number of points,
        # we interpolate the feature value for indices delineating the range of bins.
        # I.e., sorting the features values along one axis, the (xp, fp) pairs are formed where
        # xp denote the integer index of `n_samples points and fp the corresponding feature values.
        # now, the value of feature is interpolated for the indices corresponding to lower and
        # upper ranges of bins denoted by x. The x indices are obtained by evenly dividing the
        # (0, n_samples) into nbins_axis bins (hence nbins_axis + 1 for number of samples indices).
        # Obviously, in this way, the bins have roughly `n_samples` points.
        # Note that the starting and ending indices of the bins are always 0 and n_samples,
        # corresponding to the minimum and maximum of the feature values.
        # The resulting bins_edge defines a monotonically increasing array of bin edges, including
        # the rightmost edge, allowing for non-uniform bin widths.
        # The bins_edge[0] and bins_edge[-1] correspond to the min and max of the feature values.
        bins_edge = np.interp(
            x=np.linspace(0, n_samples, nbins_axis + 1), xp=np.arange(n_samples), fp=np.sort(X)
        )
        # Note: alternatively, one can use x=np.linspace(0, n_samples - 1, nbins_axis + 1) so
        # that the ending index corresponds to the index of the last sample point. This would not
        # be an issue, because the numpy interpolate function np.interp has two attributes called
        # right/left, so if the value x is outside the interpolating domain, then it returns the
        # closest data point fp[-1]/fp[0], respectively. This causes the partition into bins to
        # always include the endpoints.

        # To assign samples into bins, sample features are subtracted from the bins_edge, and
        # the index of the bins_edge where the difference switches from negative to positive is
        # the bin index that the sample is assigned to. The switch from negative to positive is
        # identified by the argmax function after setting all the non-negative values to -inf.
        pt_to_bind = bins_edge - X[:, None]
        pt_to_bind[pt_to_bind >= 0.0] = -np.inf
        bin_index = np.argmax(pt_to_bind, axis=1)
        unique_bin_ids, inverse_ids = np.unique(bin_index, return_inverse=True)
        return unique_bin_ids, inverse_ids

    def get_bins_from_method(self, X):
        r"""Assign sample points to bins based on the partitioning method.

        Parameters
        ----------
        X: ndarray of shape (n_samples, n_features)
           Feature matrix of `n_samples` samples in `n_features` dimensional space.

        Returns
        -------
        bins: dict[Tuple(int), List[int]]
            Dictionary of bins where keys are the unique bin indices (that contain at least one
            sample point) and the values are the list of sample indices in that bin.
        """

        # dictionary of bins where the keys are the unique bin indices (that contain at least one
        # sample) and the values are the list of sample indices in that bin.
        bins = {}

        if self.bin_method == "equisized_independent":
            # partition each dimension/feature independently into `num_bins_axis` bins
            unique_bin_index, inverse_index = self.partition_points_to_bins_equisized(
                X, self.nbins_axis
            )
            # populate bins dictionary
            for i, key in enumerate(unique_bin_index):
                bins[tuple(key)] = list(np.where(inverse_index == i)[0])

        elif self.bin_method == "equisized_dependent":
            # partition the first dimension (1st feature axis) into `num_bins_axis` bins
            unique_bin_index, inverse_index = self.partition_points_to_bins_equisized(
                X[:, 0], self.nbins_axis
            )
            # populate bins dictionary based on the 1st feature
            for i, key in enumerate(unique_bin_index):
                bins[tuple([key])] = list(np.where(inverse_index == i)[0])

            # loop over the remaining dimensions (2nd to last feature axis), and for each axis
            # partition the points in each bin of the previous axes into `num_bins_axis` bins
            # as a result, each iteration adds a new dimension to the bins dictionary
            for index_feature in range(1, X.shape[1]):
                # make a dictionary to store the bins for the current axis
                bins_axis = {}
                # divide points in each bin into `num_bins_axis` bins based on the i-th feature
                for bin, index_samples in bins.items():
                    # equisized partition of points in bin along i-th feature
                    unique_bin_index, inverse_index = self.partition_points_to_bins_equisized(
                        X[index_samples, index_feature], self.nbins_axis
                    )
                    # update the bins_axis to include the new dimension/feature for the current bin
                    for i, bin_index in enumerate(unique_bin_index):
                        # form a new bin_index by appending  current bin_index to the previous bin
                        key = tuple(list(bin) + [bin_index])
                        bins_axis.update(
                            {key: list(np.array(index_samples)[np.where(inverse_index == i)[0]])}
                        )
                bins = bins_axis

        elif self.bin_method == "equifrequent_independent":
            # partition each dimension of feature space independently into `num_bins_axis` bins
            bins_features = np.zeros(X.shape, dtype=int)
            for index_feature in range(0, X.shape[1]):
                unique_bin_index, inverse_index = self.partition_points_to_bins_equifrequent(
                    X[:, index_feature], self.nbins_axis
                )
                bins_features[:, index_feature] = unique_bin_index[inverse_index]
            unique_bin_index, inverse_index = np.unique(bins_features, return_inverse=True, axis=0)

            # populate bins dictionary
            for i, key in enumerate(unique_bin_index):
                bins[tuple(key)] = list(np.where(inverse_index == i)[0])

        elif self.bin_method == "equifrequent_dependent":
            # partition the first dimension (1st feature axis) into `num_bins_axis` bins
            unique_bin_index, inverse_index = self.partition_points_to_bins_equifrequent(
                X[:, 0], self.nbins_axis
            )
            # populate bins dictionary based on the 1st feature
            for i, key in enumerate(unique_bin_index):
                bins[tuple([key])] = list(np.where(inverse_index == i)[0])

            # loop over the remaining dimensions (2nd to last feature axis), and for each axis
            # partition the points in each bin of the previous axes into `num_bins_axis` bins
            for index_feature in range(1, X.shape[1]):
                # make a dictionary to store the bins for the current axis
                bins_axis = {}
                # divide points in each bin, based on the i-th feature, into `num_bins_axis` bins
                for bin, index_samples in bins.items():
                    # equifrequent partition of points in bin along i-th feature
                    unique_bin_index, inverse_index = self.partition_points_to_bins_equifrequent(
                        X[index_samples, index_feature], self.nbins_axis
                    )
                    # update the bins_axis to include the new dimension/feature for the current bin
                    for i, key in enumerate(unique_bin_index):
                        # form a new bin_index by appending  current bin_index to the previous bin
                        key = tuple(list(bin) + [key])
                        bins_axis.update(
                            {key: list(np.array(index_samples)[np.where(inverse_index == i)[0]])}
                        )
                bins = bins_axis
        else:
            raise ValueError(f"{self.bin_method} not a valid bin_method")

        return bins

    def select_from_bins(
        self,
        X,
        bins,
        num_selected,
        diversity_type="hypersphere_overlap",
        cs=None,
    ):
        r"""
        From the bins, select a certain number of points of the bins.

        Points are selected in an iterative manner. If the number of points needed to be selected
         is greater than number of bins left then randomly select points from each of the bins. If
         any of the bins are empty then remove the bins. If it is less than the number of bins left,
         then calculate the diversity of each bin and choose points of bins with the highest diversity.

        Parameters
        ----------
        X: ndarray of shape (n_samples, n_features)
            Feature matrix of `n_samples` samples in `n_features` dimensional space.
        bins: dict(tuple(int), list[int])
            The bins that map to the id the bin (as a tuple of integers) and returns
            the indices of the points that are contained in that bin.
        num_selected: int
            Number of points to select from the bins.
        diversity_type: str, optional
            Type of diversity to use. Default="hypersphere_overlap".
        cs : int, optional
            Number of common substructures in molecular compound dataset. Used only if calculating
            `explicit_diversity_index`. Default is "None".

        Returns
        -------
        List[int]:
            Indices of the points that were selected.
        """
        old_len = 0
        to_delete = []
        selected = []
        rng = np.random.default_rng(seed=self.random_seed)
        while len(selected) < num_selected:
            num_needed = num_selected - len(selected)
            # if the number of samples that should be selected is greater than number of bins,
            # randomly select points from the bins.
            if len(bins) <= num_needed:
                # Go through each bin and select a point at random from it and delete it later
                for bin_idx, bin_list in bins.items():
                    random_int = rng.integers(low=0, high=len(bin_list), size=1)[0]
                    sample_index = bin_list.pop(random_int)
                    selected.append(sample_index)
                    if len(bin_list) == 0:
                        to_delete.append(bin_idx)
                for idx in to_delete:
                    del bins[idx]
                to_delete = []
            else:
                # If number of samples that should be selected is less than the number of bins,
                # calculate the diversity of each bin and select samples from bins with highest
                # diversity.
                diversity = [
                    (
                        compute_diversity(
                            features=X,
                            feature_subset=X[bin_list, :],
                            div_type=diversity_type,
                            cs=cs,
                        ),
                        bin_idx,
                    )
                    for bin_idx, bin_list in bins.items()
                ]
                diversity.sort(reverse=True)
                for _, bin_idx in diversity[:num_needed]:
                    random_int = rng.integers(low=0, high=len(bins[bin_idx]), size=1)[0]
                    sample_index = bins[bin_idx].pop(random_int)
                    selected.append(sample_index)
            if len(selected) == old_len:
                break
            old_len = len(selected)
        return selected

    def select_from_cluster(self, X: np.ndarray, num_selected: int, cluster_ids: np.ndarray = None):
        """
        Grid partitioning algorithm for selecting points from cluster.

        Parameters
        ----------
        X: ndarray of shape (n_samples, n_features)
            Feature matrix of `n_samples` samples in `n_features` dimensional space.
        num_selected: int
            Number of molecules that need to be selected.
        cluster_ids: ndarray
            Indices of molecules that form a cluster

        Returns
        -------
        selected: list[int]
            List of ids of selected molecules with size `num_selected`.
        """
        if not isinstance(X, np.ndarray):
            raise TypeError(f"X {type(X)} should of type numpy array.")
        if not isinstance(num_selected, int):
            raise TypeError(f"num_selected {type(num_selected)} should be of type int.")
        if cluster_ids is not None and not isinstance(cluster_ids, np.ndarray):
            raise TypeError(
                f"cluster_ids {type(cluster_ids)} should be either None or numpy " f"array."
            )

        if cluster_ids is not None:
            X = X[cluster_ids]
        bins = self.get_bins_from_method(X)
        selected = self.select_from_bins(X, bins, num_selected)
        return selected


class Medoid(SelectionBase):
    """Selecting points using an algorithm adapted from KDTree.

    Points are initially used to construct a KDTree. Euclidean distances are used for this
    algorithm. The first point selected is based on the ref_index provided and becomes the first
    query point. An approximation of the furthest point to the query point is found using
    find_furthest_neighbor and is selected. find_nearest_neighbor is then done to eliminate close
    neighbors to the new selected point. Medoid is then calculated from previously selected points
    and is used as the new query point for find_furthest_neighbor, repeating the process. Terminates
    upon selecting requested number of points or if all available points exhausted.

    Adapted from: https://en.wikipedia.org/wiki/K-d_tree#Construction
    """

    def __init__(
        self,
        func_distance=lambda x, y: scipy.spatial.minkowski_distance(x, y) ** 2,
        ref_index=0,
        scaling=10,
    ):
        """
        Initializing class.

        Parameters
        ----------
        fun_distance : callable
            Function for calculating the pairwise distance between sample points.
            `fun_dist(X) -> X_dist` takes a 2D feature array of shape (n_samples, n_features)
            and returns a 2D distance array of shape (n_samples, n_samples).
        ref_index : int, optional
            Index for the sample to start selection from; this index is the first sample selected.
        scaling: float
            Percent of average maximum distance to use when eliminating the closest points.

        Notes
        -----
        The `Mediod` implementation is based on the KDTree algorithm and therefore can give
        different results for cases with duplicated points or the same features for different
        objects in the original feature space. This is dicussed in
        https://github.com/theochem/Selector/issues/238.
        This is because the same features lead to the same distances in the tree, and this is a
        known issue of sorting the points and indices in the KDTree algorithm, as discussed
        in https://github.com/scipy/scipy/issues/19029. Therefore, precautions should be taken if
        duplicated points are present in the dataset.

        """

        self.starting_idx = ref_index
        self.func_distance = func_distance
        self.BT = collections.namedtuple("BT", ["value", "index", "left", "right"])
        self.FNRecord = collections.namedtuple("FNRecord", ["point", "index", "distance"])
        self.scaling = scaling / 100
        self.ratio = None

    def _kdtree(self, arr):
        """Construct a k-d tree from an iterable of points.

        Parameters
        ----------
        arr: list or np.ndarray
            Coordinate array of points.

        Returns
        -------
        kdtree: collections.namedtuple
            KDTree organizing coordinates.
        """

        k = len(arr[0])

        def build(points, depth, old_indices=None):
            """Build a k-d tree from a set of points at a given depth."""
            if len(points) == 0:
                return None
            middle = len(points) // 2

            # sort the points and indices
            # indices, points = zip(*sorted(enumerate(points), key=lambda x: x[1][depth % k]))
            indices = np.argsort(np.array(points)[:, depth % k], kind="stable")
            points = np.array(points)[indices]

            if old_indices is not None:
                indices = [old_indices[i] for i in indices]
            return self.BT(
                value=points[middle],
                index=indices[middle],
                left=build(
                    points=points[:middle],
                    depth=depth + 1,
                    old_indices=indices[:middle],
                ),
                right=build(
                    points=points[middle + 1 :],
                    depth=depth + 1,
                    old_indices=indices[middle + 1 :],
                ),
            )

        kdtree = build(points=arr, depth=0)
        return kdtree

    def _eliminate(self, tree, point, threshold, num_eliminate, bv):
        """Eliminates points from being selected in future rounds.

        Parameters
        ----------
        tree: scipy.spatial.KDTree
            KDTree organizing coordinates.
        point: list
            Point where close neighbors should be eliminated.
        threshold: float
            An average of all the furthest distances found using find_furthest_neighbor
        num_eliminate: int
            Maximum number of points permitted to be eliminated.
        bv: bitarray
            Bitvector marking picked/eliminated points.

        Returns
        -------
        num_eliminate: int
            Maximum number of points permitted to be eliminated.
        """
        _, elim_candidates = tree.query(
            point, k=self.ratio, distance_upper_bound=np.sqrt(threshold), workers=-1
        )
        # elim_candidates can be integer or array of integers
        # https://github.com/scipy/scipy/blob/a2a287d1f7c81154256ba742b4b8bb108a612166/scipy/spatial/_kdtree.py#L476
        if isinstance(elim_candidates, np.intp):
            elim_candidates = [elim_candidates]

        if num_eliminate < 0:
            elim_candidates = elim_candidates[:num_eliminate]
        for index in elim_candidates:
            try:
                bv[index] = 1
                num_eliminate -= 1
            except IndexError:
                break
        return num_eliminate

    def _find_furthest_neighbor(self, kdtree, point, selected_bitvector):
        """Find approximately the furthest neighbor in a k-d tree for a given point.

        Parameters
        ----------
        kdtree: collections.namedtuple
            KDTree organizing coordinates.
        point: list
            Query point for search.
        selected_bitvector: bitarray
            Bitvector to keep track of previously selected points from array.

        Returns
        -------
        best: collections.namedtuple
            The furthest point found in search.
        """

        k = len(point)
        best = None

        def search(tree, depth):
            # Recursively search through the k-d tree to find the
            # furthest neighbor.

            nonlocal selected_bitvector
            nonlocal best

            if tree is None:
                return

            if not selected_bitvector[tree.index]:
                distance = self.func_distance(tree.value, point)
                if best is None or distance > best.distance:
                    best = self.FNRecord(point=tree.value, index=tree.index, distance=distance)

            axis = depth % k
            diff = point[axis] - tree.value[axis]
            if diff <= 0:
                close, away = tree.left, tree.right
            else:
                close, away = tree.right, tree.left

            search(tree=away, depth=depth + 1)
            if best is None or (
                close is not None and diff**2 <= 1.1 * ((point[axis] - close.value[axis]) ** 2)
            ):
                search(tree=close, depth=depth + 1)

        search(tree=kdtree, depth=0)
        return best

    def select_from_cluster(self, arr, num_selected, cluster_ids=None):
        """Main function for selecting points using the KDTree algorithm.

        Parameters
        ----------
        arr: np.ndarray
            Coordinate array of points
        num_selected: int
            Number of molecules that need to be selected.
        cluster_ids: np.ndarray
            Indices of molecules that form a cluster

        Returns
        -------
        selected: list
            List of ids of selected molecules
        """
        if cluster_ids is not None:
            arr = arr[cluster_ids]

        if isinstance(arr, np.ndarray):
            arr = arr.tolist()
        arr_len = len(arr)
        fartree = self._kdtree(arr)
        neartree = scipy.spatial.KDTree(arr)

        bv = bitarray.bitarray(arr_len)
        bv[:] = 0
        selected = [self.starting_idx]
        query_point = arr[self.starting_idx]
        bv[self.starting_idx] = 1
        count = 1
        num_eliminate = arr_len - num_selected
        self.ratio = math.ceil(num_eliminate / num_selected)
        best_distance_av = 0
        while len(selected) < num_selected:
            new_point = self._find_furthest_neighbor(fartree, query_point, bv)
            if new_point is None:
                return selected
            selected.append(new_point.index)
            bv[new_point.index] = 1
            query_point = (count * np.array(query_point) + np.array(new_point.point)) / (count + 1)
            query_point = query_point.tolist()
            if count == 1:
                best_distance_av = new_point.distance
            else:
                best_distance_av = (count * best_distance_av + new_point.distance) / (count + 1)
            if count == 1:
                if num_eliminate > 0 and self.scaling != 0:
                    num_eliminate = self._eliminate(
                        neartree,
                        arr[self.starting_idx],
                        best_distance_av * self.scaling,
                        num_eliminate,
                        bv,
                    )
            if num_eliminate > 0 and self.scaling != 0:
                num_eliminate = self._eliminate(
                    neartree, new_point.point, best_distance_av * self.scaling, num_eliminate, bv
                )
            count += 1
        return selected



# -*- coding: utf-8 -*-
#
# The Selector is a Python library of algorithms for selecting diverse
# subsets of data for machine-learning.
#
# Copyright (C) 2022-2024 The QC-Devs Community
#
# This file is part of Selector.
#
# Selector is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Selector is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Common functions for test module."""

from importlib import resources
from typing import Any, Tuple, Union

import numpy as np
from sklearn.datasets import make_blobs
from sklearn.metrics import pairwise_distances

__all__ = [
    "generate_synthetic_cluster_data",
    "generate_synthetic_data",
    "get_data_file_path",
]


def generate_synthetic_cluster_data():
    # generate the first cluster with 3 points
    cluster_one = np.array([[0, 0], [0, 1], [0, 2]])
    # generate the second cluster with 6 points
    cluster_two = np.array([[3, 0], [3, 1], [3, 2], [3, 3], [3, 4], [3, 5]])
    # generate the third cluster with 9 points
    cluster_three = np.array(
        [[6, 0], [6, 1], [6, 2], [6, 3], [6, 4], [6, 5], [6, 6], [6, 7], [6, 8]]
    )
    # concatenate the clusters
    coords = np.vstack([cluster_one, cluster_two, cluster_three])
    # generate the labels
    labels = np.hstack([[0 for _ in range(3)], [1 for _ in range(6)], [2 for _ in range(9)]])

    return coords, labels, cluster_one, cluster_two, cluster_three


def generate_synthetic_data(
    n_samples: int = 100,
    n_features: int = 2,
    n_clusters: int = 2,
    cluster_std: float = 1.0,
    center_box: Tuple[float, float] = (-10.0, 10.0),
    metric: str = "euclidean",
    shuffle: bool = True,
    random_state: int = 42,
    pairwise_dist: bool = False,
    **kwargs: Any,
) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray, np.ndarray]]:
    """Generate synthetic data.

    Parameters
    ----------
    n_samples : int, optional
        The number of sample points.
    n_features : int, optional
        The number of features.
    n_clusters : int, optional
        The number of clusters.
    cluster_std : float, optional
        The standard deviation of the clusters.
    center_box : tuple[float, float], optional
        The bounding box for each cluster center when centers are generated at random.
    metric : str, optional
        The metric used for computing pairwise distances. For the supported
        distance matrix, please refer to
        https://scikit-learn.org/stable/modules/generated/sklearn.metrics.pairwise_distances.html.
    shuffle : bool, optional
        Whether to shuffle the samples.
    random_state : int, optional
        The random state used for generating synthetic data.
    pairwise_dist : bool, optional
        If True, then compute and return the pairwise distances between sample points.
    **kwargs : Any, optional
            Additional keyword arguments for the scikit-learn `pairwise_distances` function.

    Returns
    -------
    syn_data : np.ndarray
        The synthetic data.
    class_labels : np.ndarray
        The integer labels for cluster membership of each sample.
    dist: np.ndarray
        The symmetric pairwise distances between samples.

    """
    # pylint: disable=W0632
    syn_data, class_labels = make_blobs(
        n_samples=n_samples,
        n_features=n_features,
        centers=n_clusters,
        cluster_std=cluster_std,
        center_box=center_box,
        shuffle=shuffle,
        random_state=random_state,
        return_centers=False,
    )
    if pairwise_dist:
        dist = pairwise_distances(
            X=syn_data,
            Y=None,
            metric=metric,
            **kwargs,
        )
        return syn_data, class_labels, dist
    return syn_data, class_labels


def get_data_file_path(file_name):
    """Get the absolute path of the data file inside the package.

    Parameters
    ----------
    file_name : str
        The name of the data file to load.

    Returns
    -------
    str
        The absolute path of the data file inside the package

    """
    data_file_path = resources.files("selector.methods.tests").joinpath(f"data/{file_name}")

    return data_file_path



# -*- coding: utf-8 -*-
#
# The Selector is a Python library of algorithms for selecting diverse
# subsets of data for machine-learning.
#
# Copyright (C) 2022-2024 The QC-Devs Community
#
# This file is part of Selector.
#
# Selector is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Selector is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Test Converter Module."""

import numpy as np
from numpy.testing import assert_almost_equal, assert_equal, assert_raises

from selector.measures import converter as cv

# Tests for variations on input `x` for sim_to_dist()


def test_sim_2_dist_float_int():
    """Test similarity to distance input handling when input is a float or int."""
    expected_1 = 0.25
    int_out = cv.sim_to_dist(4, "reciprocal")
    assert_equal(int_out, expected_1)
    expected_2 = 2
    float_out = cv.sim_to_dist(0.5, "reciprocal")
    assert_equal(float_out, expected_2)


def test_sim_2_dist_array_dimension_error():
    """Test sim to dist function with incorrect input dimensions for `x`."""
    assert_raises(ValueError, cv.sim_to_dist, np.ones([2, 2, 2]), "reciprocal")


def test_sim_2_dist_1d_metric_error():
    """Test sim to dist function with an invalid metric for 1D arrays."""
    assert_raises(ValueError, cv.sim_to_dist, np.ones(5), "gravity")
    assert_raises(ValueError, cv.sim_to_dist, np.ones(5), "co-occurrence")


# Tests for variations on input `metric` for sim_to_dist()


def test_sim_2_dist():
    """Test similarity to distance method with specified metric."""
    x = np.array([[1, 0.2, 0.5], [0.2, 1, 0.25], [0.5, 0.25, 1]])
    expected = np.array([[0.20, 1, 0.70], [1, 0.20, 0.95], [0.70, 0.95, 0.20]])
    actual = cv.sim_to_dist(x, "reverse")
    assert_almost_equal(actual, expected, decimal=10)


def test_sim_2_dist_frequency():
    """Test similarity to distance method with a frequency metric."""
    x = np.array([[4, 9, 1], [9, 1, 25], [1, 25, 16]])
    expected = np.array([[(1 / 2), (1 / 3), 1], [(1 / 3), 1, (1 / 5)], [1, (1 / 5), (1 / 4)]])
    actual = cv.sim_to_dist(x, "transition")
    assert_almost_equal(actual, expected, decimal=10)


def test_sim_2_dist_frequency_error():
    """Test similarity to distance method with a frequency metric and incorrect input."""
    # zeroes in the frequency matrix
    x = np.array([[0, 9, 1], [9, 1, 25], [1, 25, 0]])
    assert_raises(ValueError, cv.sim_to_dist, x, "gravity")
    # negatives in the frequency matrix
    y = np.array([[1, -9, 1], [9, 1, -25], [1, 25, 16]])
    assert_raises(ValueError, cv.sim_to_dist, x, "gravity")


def test_sim_2_dist_membership():
    """Test similarity to distance method with the membership metric."""
    # x = np.array([[(1 / 2), (1 / 5)], [(1 / 4), (1 / 3)]])
    x = np.array([[1, 1 / 5, 1 / 3], [1 / 5, 1, 4 / 5], [1 / 3, 4 / 5, 1]])
    expected = np.array([[0, 4 / 5, 2 / 3], [4 / 5, 0, 1 / 5], [2 / 3, 1 / 5, 0]])
    actual = cv.sim_to_dist(x, "membership")
    assert_almost_equal(actual, expected, decimal=10)


def test_sim_2_dist_membership_error():
    """Test similarity to distance method with the membership metric when there is an input error."""
    x = np.array([[1, 0, -7], [0, 1, 3], [-7, 3, 1]])
    assert_raises(ValueError, cv.sim_to_dist, x, "membership")


def test_sim_2_dist_invalid_metric():
    """Test similarity to distance method with an unsupported metric."""
    assert_raises(ValueError, cv.sim_to_dist, np.ones(5), "testing")


def test_sim_2_dist_non_symmetric():
    """Test the invalid 2D symmetric matrix error."""
    x = np.array([[1, 2], [4, 5]])
    assert_raises(ValueError, cv.sim_to_dist, x, "reverse")


# Tests for individual metrics


def test_reverse():
    """Test the reverse function for similarity to distance conversion."""
    x = np.array([[3, 1, 1], [1, 3, 0], [1, 0, 3]])
    expected = np.array([[0, 2, 2], [2, 0, 3], [2, 3, 0]])
    actual = cv.reverse(x)
    assert_equal(actual, expected)


def test_reciprocal():
    """Test the reverse function for similarity to distance conversion."""
    x = np.array([[1, 0.25, 0.40], [0.25, 1, 0.625], [0.40, 0.625, 1]])
    expected = np.array([[1, 4, 2.5], [4, 1, 1.6], [2.5, 1.6, 1]])
    actual = cv.reciprocal(x)
    assert_equal(actual, expected)


def test_reciprocal_error():
    """Test the reverse function with incorrect input values."""
    # zero value for similarity (causes divide by zero issues)
    x = np.array([[0, 4], [3, 2]])
    assert_raises(ValueError, cv.reciprocal, x)
    # negative value for similarity (distance cannot be negative)
    y = np.array([[1, -4], [3, 2]])
    assert_raises(ValueError, cv.reciprocal, y)


def test_exponential():
    """Test the exponential function for similarity to distance conversion."""
    x = np.array([[1, 0.25, 0.40], [0.25, 1, 0.625], [0.40, 0.625, 1]])
    expected = np.array(
        [
            [0, 1.38629436112, 0.91629073187],
            [1.38629436112, 0, 0.47000362924],
            [0.91629073187, 0.47000362924, 0],
        ]
    )
    actual = cv.exponential(x)
    assert_almost_equal(actual, expected, decimal=10)


def test_exponential_error():
    """Test the exponential function when max similarity is zero."""
    x = np.zeros((4, 4))
    assert_raises(ValueError, cv.exponential, x)


def test_gaussian():
    """Test the gaussian function for similarity to distance conversion."""
    x = np.array([[1, 0.25, 0.40], [0.25, 1, 0.625], [0.40, 0.625, 1]])
    expected = np.array(
        [
            [0, 1.17741002252, 0.95723076208],
            [1.17741002252, 0, 0.68556810693],
            [0.95723076208, 0.68556810693, 0],
        ]
    )
    actual = cv.gaussian(x)
    assert_almost_equal(actual, expected, decimal=10)


def test_gaussian_error():
    """Test the gaussian function when max similarity is zero."""
    x = np.zeros((4, 4))
    assert_raises(ValueError, cv.gaussian, x)


def test_correlation():
    """Test the correlation to distance conversion function."""
    x = np.array([[1, 0.5, 0.2], [0.5, 1, -0.2], [0.2, -0.2, 1]])
    # expected = sqrt(1-x)
    expected = np.array(
        [
            [0, 0.70710678118, 0.894427191],
            [0.70710678118, 0, 1.09544511501],
            [0.894427191, 1.09544511501, 0],
        ]
    )
    actual = cv.correlation(x)
    assert_almost_equal(actual, expected, decimal=10)


def test_correlation_error():
    """Test the correlation function with an out of bounds array."""
    x = np.array([[1, 0, -7], [0, 1, 3], [-7, 3, 1]])
    assert_raises(ValueError, cv.correlation, x)


def test_transition():
    """Test the transition function for frequency to distance conversion."""
    x = np.array([[4, 9, 1], [9, 1, 25], [1, 25, 16]])
    expected = np.array([[(1 / 2), (1 / 3), 1], [(1 / 3), 1, (1 / 5)], [1, (1 / 5), (1 / 4)]])

    actual = cv.transition(x)
    assert_almost_equal(actual, expected, decimal=10)


def test_co_occurrence():
    """Test the co-occurrence conversion function."""
    x = np.array([[1, 2, 3], [2, 1, 3], [3, 3, 1]])
    expected = np.array(
        [
            [1 / (19 / 121 + 1), 1 / (38 / 121 + 1), 1 / (57 / 121 + 1)],
            [1 / (38 / 121 + 1), 1 / (19 / 121 + 1), 1 / (57 / 121 + 1)],
            [1 / (57 / 121 + 1), 1 / (57 / 121 + 1), 1 / (19 / 121 + 1)],
        ]
    )
    actual = cv.co_occurrence(x)
    assert_almost_equal(actual, expected, decimal=10)


def test_gravity():
    """Test the gravity conversion function."""
    x = np.array([[1, 2, 3], [2, 1, 3], [3, 3, 1]])
    expected = np.array(
        [
            [2.5235730726, 1.7844356324, 1.45698559277],
            [1.7844356324, 2.5235730726, 1.45698559277],
            [1.45698559277, 1.45698559277, 2.5235730726],
        ]
    )
    actual = cv.gravity(x)
    assert_almost_equal(actual, expected, decimal=10)


def test_probability():
    """Test the probability to distance conversion function."""
    x = np.array([[0.3, 0.7], [0.5, 0.5]])
    expected = np.array([[1.8116279322, 1.1356324735], [1.3819765979, 1.3819765979]])
    actual = cv.probability(x)
    assert_almost_equal(actual, expected, decimal=10)


def test_probability_error():
    """Test the correlation function with an out of bounds array."""
    # negative value for probability
    x = np.array([[-0.5]])
    assert_raises(ValueError, cv.probability, x)
    # too large value for probability
    y = np.array([[3]])
    assert_raises(ValueError, cv.probability, y)
    # zero value for probability (causes divide by zero issues)
    z = np.array([[0]])
    assert_raises(ValueError, cv.probability, z)


def test_covariance():
    """Test the covariance to distance conversion function."""
    x = np.array([[4, -4], [-4, 6]])
    expected = np.array([[0, 4.24264068712], [4.24264068712, 0]])
    actual = cv.covariance(x)
    assert_almost_equal(actual, expected, decimal=10)


def test_covariance_error():
    """Test the covariance function when input contains a negative variance."""
    x = np.array([[-4, 4], [4, 6]])
    assert_raises(ValueError, cv.covariance, x)
